from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from typing import ClassVar

from arti.internal.models import Model
from arti.types import Type, TypeSystem


class Format(Model):
    """Format represents file formats such as CSV, Parquet, native (eg: databases), etc.

    Formats are associated with a type system that provides a bridge between the internal
    Artigraph types and any external type information.
    """

    _abstract_ = True
    extension: str = ""
    type_system: ClassVar[TypeSystem]

    def supports(self, type_: Type) -> None:
        # TODO: Check self.type_system supports the type. We can likely add a TypeSystem method
        # that will check for matching TypeAdapters.
        pass



from __future__ import annotations

import datetime
from collections.abc import Mapping
from functools import partial
from itertools import chain
from typing import Any, Literal, Optional, TypedDict, Union, get_args, get_origin, get_type_hints

import arti.types
from arti.internal.type_hints import (
    NoneType,
    is_optional_hint,
    is_typeddict,
    is_union,
    lenient_issubclass,
)
from arti.types import Type, TypeAdapter, TypeSystem, _ScalarClassTypeAdapter

python_type_system = TypeSystem(key="python")
_generate = partial(_ScalarClassTypeAdapter.generate, type_system=python_type_system)

# NOTE: issubclass(bool, int) is True, so set higher priority
_generate(artigraph=arti.types.Boolean, system=bool, priority=int(1e9))
_generate(artigraph=arti.types.Date, system=datetime.date)
_generate(artigraph=arti.types.Null, system=NoneType)
_generate(artigraph=arti.types.String, system=str)
for _precision in (16, 32, 64):
    _generate(
        artigraph=getattr(arti.types, f"Float{_precision}"),
        system=float,
        priority=_precision,
    )
for _precision in (8, 16, 32, 64):
    _generate(
        artigraph=getattr(arti.types, f"Int{_precision}"),
        system=int,
        priority=_precision,
    )


@python_type_system.register_adapter
class PyDatetime(_ScalarClassTypeAdapter):
    artigraph = arti.types.Timestamp
    system = datetime.datetime

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> Type:
        return cls.artigraph(precision="microsecond")


class PyValueContainer(TypeAdapter):
    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> Type:
        (element,) = get_args(type_)
        return cls.artigraph(
            element=python_type_system.to_artigraph(element, hints=hints),
        )

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        return lenient_issubclass(get_origin(type_), cls.system)

    @classmethod
    def to_system(cls, type_: Type, *, hints: dict[str, Any]) -> Any:
        assert isinstance(type_, cls.artigraph)
        return cls.system[
            python_type_system.to_system(type_.element, hints=hints),  # type: ignore
        ]


@python_type_system.register_adapter
class PyList(PyValueContainer):
    artigraph = arti.types.List
    system = list
    priority = 1


# NOTE: PyTuple only covers sequences (eg: tuple[int, ...]), not structure (eg: tuple[int, str]).
@python_type_system.register_adapter
class PyTuple(PyValueContainer):
    artigraph = arti.types.List
    system = tuple

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> Type:
        origin, args = get_origin(type_), get_args(type_)
        assert origin is not None
        assert len(args) == 2 and args[1] is ...
        return super().to_artigraph(origin[args[0]], hints=hints)

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        if super().matches_system(type_, hints=hints):
            args = get_args(type_)
            if len(args) == 2 and args[1] is ...:
                return True
        return False

    @classmethod
    def to_system(cls, type_: Type, *, hints: dict[str, Any]) -> Any:
        ret = super().to_system(type_, hints=hints)
        origin, args = get_origin(ret), get_args(ret)
        assert origin is not None
        assert len(args) == 1
        return origin[args[0], ...]


@python_type_system.register_adapter
class PyFrozenset(PyValueContainer):
    artigraph = arti.types.Set
    system = frozenset


@python_type_system.register_adapter
class PySet(PyValueContainer):
    artigraph = arti.types.Set
    system = set
    priority = 1  # Set above frozenset


@python_type_system.register_adapter
class PyLiteral(TypeAdapter):
    artigraph = arti.types.Enum
    system = Literal

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> Type:
        origin, items = get_origin(type_), get_args(type_)
        if is_union(origin):
            assert not is_optional_hint(type_)  # Should be handled by PyOptional
            # We only support Enums currently, so all subtypes must be Literal
            if non_literals := [sub for sub in items if not get_origin(sub) is Literal]:
                raise NotImplementedError(
                    f"Only Union[Literal[...], ...] (enums) are currently supported, got invalid subtypes: {non_literals}"
                )
            # Flatten Union[Literal[1], Literal[1,2,3]]
            origin, items = Literal, tuple(chain.from_iterable(get_args(sub) for sub in items))
        assert origin is Literal
        assert isinstance(items, tuple)
        if len(items) == 0:
            raise NotImplementedError(f"Invalid Literal with no values: {type_}")
        py_type, *other_types = [type(v) for v in items]
        if not all(t is py_type for t in other_types):
            raise ValueError("All Literals must be the same type, got: {(py_type, *other_types)}")
        return cls.artigraph(
            type=python_type_system.to_artigraph(py_type, hints=hints),
            items=items,
        )

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        # We don't (currently) support arbitrary Unions, but can map Union[Literal[1], Literal[2]]
        # to an Enum. Python's Optional is also represented as a Union, but we handle that with the
        # high priority PyOptional.
        origin, items = get_origin(type_), get_args(type_)
        return origin is Literal or (
            is_union(origin) and all(get_origin(sub) is Literal for sub in items)
        )

    @classmethod
    def to_system(cls, type_: Type, *, hints: dict[str, Any]) -> Any:
        assert isinstance(type_, cls.artigraph)
        return cls.system[tuple(type_.items)]


@python_type_system.register_adapter
class PyMap(TypeAdapter):
    artigraph = arti.types.Map
    system = dict

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> Type:
        key, value = get_args(type_)
        return cls.artigraph(
            key=python_type_system.to_artigraph(key, hints=hints),
            value=python_type_system.to_artigraph(value, hints=hints),
        )

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        return lenient_issubclass(get_origin(type_), (cls.system, Mapping))

    @classmethod
    def to_system(cls, type_: Type, *, hints: dict[str, Any]) -> Any:
        assert isinstance(type_, cls.artigraph)
        return cls.system[
            python_type_system.to_system(type_.key, hints=hints),
            python_type_system.to_system(type_.value, hints=hints),
        ]  # type: ignore


@python_type_system.register_adapter
class PyOptional(TypeAdapter):
    artigraph = arti.types.Type  # Check against isinstance *and* .nullable
    system = Optional
    # Set very high priority to intercept other matching arti.types.Types/py Union in order to set .nullable
    priority = int(1e9)

    @classmethod
    def matches_artigraph(cls, type_: Type, *, hints: dict[str, Any]) -> bool:
        return super().matches_artigraph(type_, hints=hints) and type_.nullable

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> Type:
        # Optional is represented as a Union; strip out NoneType before dispatching
        type_ = Union[tuple(subtype for subtype in get_args(type_) if subtype is not NoneType)]
        return python_type_system.to_artigraph(type_, hints=hints).copy(update={"nullable": True})

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        return is_optional_hint(type_)

    @classmethod
    def to_system(cls, type_: Type, *, hints: dict[str, Any]) -> Any:
        return cls.system[
            python_type_system.to_system(type_.copy(update={"nullable": False}), hints=hints)
        ]


@python_type_system.register_adapter
class PyStruct(TypeAdapter):
    artigraph = arti.types.Struct
    system = TypedDict

    # TODO: Support and inspect TypedDict's '__optional_keys__', '__required_keys__', '__total__'

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> Type:
        return arti.types.Struct(
            name=type_.__name__,
            fields={
                field_name: python_type_system.to_artigraph(field_type, hints=hints)
                for field_name, field_type in get_type_hints(type_).items()
            },
        )

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        # NOTE: This check is probably a little shaky, particularly across python versions. Consider
        # using the typing_inspect package.
        return is_typeddict(type_)

    @classmethod
    def to_system(cls, type_: Type, *, hints: dict[str, Any]) -> Any:
        assert isinstance(type_, cls.artigraph)
        return TypedDict(
            type_.name,
            {
                field_name: python_type_system.to_system(field_type, hints=hints)
                for field_name, field_type in type_.fields.items()
            },
        )  # type: ignore



from datetime import date, datetime

from arti.internal.type_hints import NoneType
from arti.types.python import python_type_system
from arti.views import View


class PythonBuiltin(View):
    _abstract_ = True

    type_system = python_type_system


class Date(PythonBuiltin):
    python_type = date


class Datetime(PythonBuiltin):
    python_type = datetime


class Dict(PythonBuiltin):
    python_type = dict


class Float(PythonBuiltin):
    python_type = float


class Int(PythonBuiltin):
    python_type = int


class List(PythonBuiltin):
    python_type = list


class Null(PythonBuiltin):
    python_type = NoneType


class Str(PythonBuiltin):
    python_type = str



from arti.formats import Format
from arti.types.python import python_type_system


class JSON(Format):
    extension: str = "json"
    # Perhaps we narrow down a json_type_system with the subset of supported types + a way to hook
    # into json.JSON{De,En}coders.
    type_system = python_type_system



from __future__ import annotations

import hashlib
import tempfile
from glob import glob
from pathlib import Path

from arti.fingerprints import Fingerprint
from arti.partitions import CompositeKeyTypes
from arti.storage import InputFingerprints, Storage, StoragePartition
from arti.storage._internal import parse_spec, spec_to_wildcard


class LocalFilePartition(StoragePartition):
    path: str

    def compute_fingerprint(self, buffer_size: int = 1024 * 1024) -> Fingerprint:
        with open(self.path, mode="rb") as f:
            sha = hashlib.sha1()
            data = f.read(buffer_size)
            while len(data) > 0:
                sha.update(data)
                data = f.read(buffer_size)
        return Fingerprint.from_string(sha.hexdigest())


class LocalFile(Storage[LocalFilePartition]):
    path: str = str(
        Path(tempfile.gettempdir())
        / "{graph_name}"
        / "{path_tags}"
        / "{names}"
        / "{partition_key_spec}"
        / "{input_fingerprint}"
        / "{name}.{extension}"
    )

    def discover_partitions(
        self,
        key_types: CompositeKeyTypes,
        input_fingerprints: InputFingerprints = InputFingerprints(),
    ) -> tuple[LocalFilePartition, ...]:
        wildcard = spec_to_wildcard(self.path, key_types)
        paths = set(glob(wildcard))
        path_metadata = parse_spec(
            paths, spec=self.path, key_types=key_types, input_fingerprints=input_fingerprints
        )
        return tuple(
            self.generate_partition(
                keys=keys,
                input_fingerprint=input_fingerprint,
            )
            for path, (input_fingerprint, keys) in path_metadata.items()
        )



from typing import Any, Protocol

from pydantic import BaseModel
from pydantic.fields import ModelField
from pydantic.fields import UndefinedType as _PydanticUndefinedType

from arti.internal.type_hints import lenient_issubclass
from arti.types import Struct, Type, TypeAdapter, TypeSystem, _ScalarClassTypeAdapter
from arti.types.python import python_type_system

pydantic_type_system = TypeSystem(key="pydantic")


class _PostFieldConversionHook(Protocol):
    def __call__(self, type_: Type, *, name: str, required: bool) -> Type:
        raise NotImplementedError()


def get_post_field_conversion_hook(type_: Any) -> _PostFieldConversionHook:
    if hasattr(type_, "_pydantic_type_system_post_field_conversion_hook_"):
        return type_._pydantic_type_system_post_field_conversion_hook_  # type: ignore
    return lambda type_, *, name, required: type_


@pydantic_type_system.register_adapter
class BaseModelAdapter(TypeAdapter):
    artigraph = Struct
    system = BaseModel

    @staticmethod
    def _field_to_artigraph(field: ModelField, *, hints: dict[str, Any]) -> Type:
        subtype = python_type_system.to_artigraph(field.outer_type_, hints=hints)
        return get_post_field_conversion_hook(subtype)(
            subtype,
            name=field.name,
            required=(
                True if isinstance(field.required, _PydanticUndefinedType) else field.required
            ),
        )

    @classmethod
    def to_artigraph(cls, type_: type[BaseModel], *, hints: dict[str, Any]) -> Type:
        return Struct(
            name=type_.__name__,
            fields={
                field.name: cls._field_to_artigraph(field, hints=hints)
                for field in type_.__fields__.values()
            },
        )

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        return lenient_issubclass(type_, cls.system)

    @classmethod
    def to_system(cls, type_: Type, *, hints: dict[str, Any]) -> type[BaseModel]:
        assert isinstance(type_, Struct)
        return type(
            f"{type_.name}",
            (BaseModel,),
            {
                "__annotations__": {
                    k: (
                        pydantic_type_system.to_system(v, hints=hints)
                        if isinstance(v, Struct)
                        else python_type_system.to_system(v, hints=hints)
                    )
                    for k, v in type_.fields.items()
                }
            },
        )


# Extend the python_type_system to handle BaseModel. This simplifies conversion of nested models


@python_type_system.register_adapter
class _PythonBaseModelAdapter(_ScalarClassTypeAdapter):
    artigraph = Struct
    system = BaseModel
    priority = int(1e8)  # Beneath the Optional Adapter

    @classmethod
    def matches_artigraph(cls, type_: Type, *, hints: dict[str, Any]) -> bool:
        # Avoid converting a python type to a BaseModel unless explicit annotated.
        return super().matches_artigraph(type_, hints=hints) and hints.get(
            f"{pydantic_type_system.key}.is_model", False
        )

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> Type:
        return BaseModelAdapter.to_artigraph(type_, hints=hints)

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        return BaseModelAdapter.matches_system(type_, hints=hints)

    @classmethod
    def to_system(cls, type_: Type, *, hints: dict[str, Any]) -> Any:
        return BaseModelAdapter.to_system(type_, hints=hints)



from __future__ import annotations

import inspect
import sys
import types
from collections.abc import Callable
from typing import _AnnotatedAlias  # type: ignore
from typing import Annotated, Any, Union, cast, get_args, get_origin, get_type_hints

NoneType = cast(type, type(None))  # mypy otherwise treats type(None) as an object


def _check_issubclass(klass: Any, check_type: type) -> bool:
    # If a hint is Annotated, we want to unwrap the underlying type and discard the rest of the
    # annotations.
    klass_origin, klass_args = get_origin(klass), get_args(klass)
    if klass_origin is Annotated:
        klass = klass_args[0]
        klass_origin, klass_args = get_origin(klass), get_args(klass)
    check_type_origin, check_type_args = get_origin(check_type), get_args(check_type)
    if check_type_origin is Annotated:
        check_type = check_type_args[0]
        check_type_origin, check_type_args = get_origin(check_type), get_args(check_type)
    # eg: issubclass(tuple, tuple)
    if klass_origin is None and check_type_origin is None:
        return issubclass(klass, check_type)
    # eg: issubclass(tuple[int], tuple)
    if klass_origin is not None and check_type_origin is None:
        return issubclass(klass_origin, check_type)
    # eg: issubclass(tuple, tuple[int])
    if klass_origin is None and check_type_origin is not None:
        return issubclass(klass, check_type_origin) and not check_type_args
    # eg: issubclass(tuple[int], tuple[int])
    if klass_origin is not None and check_type_origin is not None:
        # NOTE: Considering all container types covariant for simplicity (mypy may be more strict).
        #
        # The builtin mutable containers (list, dict, etc) are invariant (klass_args ==
        # check_type_args), but the interfaces (Mapping, Sequence, etc) and immutable containers are
        # covariant.
        if check_type_args:
            if not (
                len(klass_args) == len(check_type_args)
                and all(
                    # check subclass OR things like "..."
                    lenient_issubclass(klass_arg, check_type_arg) or klass_arg is check_type_arg
                    for (klass_arg, check_type_arg) in zip(klass_args, check_type_args)
                )
            ):
                return False
        return lenient_issubclass(klass_origin, check_type_origin)
    # Shouldn't happen, but need to explicitly say "x is not None" to narrow mypy types.
    raise NotImplementedError("The origin conditions don't cover all cases!")


def get_class_type_vars(klass: type) -> tuple[type, ...]:
    base = klass.__orig_bases__[0]  # type: ignore
    base_origin = get_origin(base)
    if base_origin is None:
        raise TypeError(f"{klass.__name__} must subclass a subscripted Generic")
    assert lenient_issubclass(klass, base_origin)
    return get_args(base)


def lenient_issubclass(klass: Any, class_or_tuple: Union[type, tuple[type, ...]]) -> bool:
    if not (isinstance(klass, type) or is_Annotated(klass)):
        return False
    if isinstance(class_or_tuple, tuple):
        return any(lenient_issubclass(klass, subtype) for subtype in class_or_tuple)
    check_type = class_or_tuple
    if is_union_hint(check_type):
        return any(lenient_issubclass(klass, subtype) for subtype in get_args(check_type))
    return _check_issubclass(klass, check_type)


def _tidy_return(return_annotation: Any, *, force_tuple_return: bool) -> Any:
    if not force_tuple_return:
        return return_annotation
    if lenient_issubclass(get_origin(return_annotation), tuple):
        return get_args(return_annotation)
    return (return_annotation,)


def tidy_signature(
    fn: Callable[..., Any],
    sig: inspect.Signature,
    *,
    force_tuple_return: bool = False,
    remove_owner: bool = False,
) -> inspect.Signature:
    type_hints = get_type_hints(fn, include_extras=True)
    sig = sig.replace(return_annotation=type_hints.get("return", sig.return_annotation))
    return sig.replace(
        parameters=[
            p.replace(annotation=type_hints.get(p.name, p.annotation))
            for p in sig.parameters.values()
            if (p.name not in ("cls", "self") if remove_owner else True)
        ],
        return_annotation=(
            sig.empty
            if sig.return_annotation is sig.empty
            else _tidy_return(sig.return_annotation, force_tuple_return=force_tuple_return)
        ),
    )


def signature(
    fn: Callable[..., Any],
    *,
    follow_wrapped: bool = True,
    force_tuple_return: bool = False,
    remove_owner: bool = True,
) -> inspect.Signature:
    """Convenience wrapper around `inspect.signature`.

    The returned Signature will have `cls`/`self` parameters removed if `remove_owner` is `True` and
    `tuple[...]` converted to `tuple(...)` in the `return_annotation`.
    """
    return tidy_signature(
        fn=fn,
        sig=inspect.signature(fn, follow_wrapped=follow_wrapped),
        force_tuple_return=force_tuple_return,
        remove_owner=remove_owner,
    )


#############################################
# Helpers for typing across python versions #
#############################################
#
# Focusing on  3.9+ (for now)


if sys.version_info < (3, 10):
    from typing import _TypedDictMeta  # type: ignore

    def is_union(type_: Any) -> bool:
        return type_ is Union

    def is_typeddict(type_: Any) -> bool:
        return isinstance(type_, _TypedDictMeta)


else:  # pragma: no cover
    from typing import is_typeddict  # noqa: F401

    def is_union(type_: Any) -> bool:
        # `Union[int, str]` or `int | str`
        return type_ is Union or type_ is types.Union


def is_Annotated(type_: Any) -> bool:
    return isinstance(type_, _AnnotatedAlias)


def is_optional_hint(type_: Any) -> bool:
    # Optional[x] is represented as Union[x, NoneType]
    return is_union(get_origin(type_)) and NoneType in get_args(type_)


def is_union_hint(type_: Any) -> bool:
    return get_origin(type_) is Union



import logging
from graphlib import TopologicalSorter
from itertools import chain

from arti.artifacts import Artifact
from arti.backends import BackendProtocol
from arti.executors import Executor
from arti.graphs import Graph
from arti.producers import Producer
from arti.storage import InputFingerprints

# TODO: Factor this code out to reusable helpers/better homes.
#   - Perhaps a lot of the _build_producer logic can live in a few Producer methods. We
#   still want to expose each partition-to-build as a parallelizable unit of execution,
#   so maybe one Producer method kicks out all to-be-built partitions+metadata, and then
#   Executor can parallelize the `.build`.


class LocalExecutor(Executor):
    def _build_artifact(self, graph: Graph, backend: BackendProtocol, artifact: Artifact) -> None:
        # NOTE: Should we do different things for "raw" vs produced artifacts?
        #
        # Eg: for raw we discover partitions, but for generated we... just
        # lookup from the Backend somehow? For all, we should probably compute
        # the statistics (or maybe compute should be part of the Producer step
        # using the exiting Views?) and check thresholds. Threshold checking
        # should apply for every run, but the statitics should no-op if already
        # computed obviously.
        #
        # We can only discover raw Artifacts - all produced ones should be written   his will fail for generated Artifacts (missing input_fingerprints). So,
        # maybe we only discover "if artifact.producer_output is None"
        if artifact.producer_output is None:
            backend.write_artifact_partitions(
                artifact,
                tuple(
                    partition
                    for partition in artifact.storage.discover_partitions(
                        artifact.partition_key_types
                    )
                ),
            )
        # TODO: Check any assumptions about what/how many should be in the db?
        # TODO: Calculate stats, run validation, etc...

    # NOTE: This should probably be broken up to separate the .map and .build (ie: so we
    # can map ahead of time and then parallelize build calls).
    def _build_producer(self, graph: Graph, backend: BackendProtocol, producer: Producer) -> None:
        input_partitions = {
            name: backend.read_artifact_partitions(artifact)
            for name, artifact in producer.inputs.items()
        }
        partition_dependencies = producer.map(
            **{
                name: partitions
                for name, partitions in input_partitions.items()
                if name in producer._map_input_metadata_
            }
        )
        # TODO: Need to validate the partition_dependencies against the Producer's
        # partitioning scheme and such (basically, check user error). eg: if output is
        # not partitioned, we expect only 1 entry in partition_dependencies
        # (NotPartitioned).
        partition_input_fingerprints = InputFingerprints(
            {
                composite_key: producer.compute_input_fingerprint(dependency_partitions)
                for composite_key, dependency_partitions in partition_dependencies.items()
            }
        )
        output_artifacts = graph.producer_outputs[producer]
        existing_output_partitions = {
            # NOTE: We'll append newly build partitions here as they are build.
            output: output.discover_storage_partitions(
                input_fingerprints=partition_input_fingerprints
            )
            for output in output_artifacts
        }
        for artifact, partitions in existing_output_partitions.items():
            backend.write_artifact_partitions(artifact, partitions)
        # TODO: Guarantee all outputs have the same set of identified partitions
        existing_output_keys = set(
            partition.keys for partition in chain.from_iterable(existing_output_partitions.values())
        )
        for output_partition_key, dependencies in partition_dependencies.items():
            if output_partition_key in existing_output_keys:
                pk_str = f" for: {dict(output_partition_key)}" if output_partition_key else "."
                logging.info(f"Skipping existing {type(producer).__name__} output{pk_str}")
                continue
            # TODO: Catch DispatchError and give a nicer error... maybe add this to our
            # @dispatch wrapper (eg: msg arg, or even fn that returns the message to
            # raise).
            arguments = {
                name: graph.read(
                    artifact=producer.inputs[name],
                    storage_partitions=partition_dependencies[output_partition_key][name],
                    # TODO: We'll probably need to update Producer._build_input_views_
                    # to hold *instances* (either objects set in Annotated or init
                    # during Producer.__init_subclass__)
                    view=view_class(),
                )
                for name, view_class in producer._build_input_views_.items()
            }
            outputs = producer.build(**arguments)
            if len(producer._output_metadata_) == 1:
                outputs = (outputs,)
            for i, output in enumerate(outputs):
                graph.write(
                    output,
                    artifact=output_artifacts[i],
                    input_fingerprint=partition_input_fingerprints[output_partition_key],
                    keys=output_partition_key,
                    view=producer._output_metadata_[i][1](),
                )

    # TODO: Support "dry_run"?
    def build(self, graph: Graph) -> None:
        with graph.backend.connect() as backend:
            for node in TopologicalSorter(graph.dependencies).static_order():
                if isinstance(node, Artifact):
                    logging.info(f"Syncing {type(node).__name__}")
                    self._build_artifact(graph, backend, node)
                elif isinstance(node, Producer):
                    logging.info(f"Building {type(node).__name__}")
                    self._build_producer(graph, backend, node)
                else:
                    raise NotImplementedError()
        logging.info(f"Build finished")



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

import abc
import os
from typing import Any, ClassVar, Generic, Optional, TypeVar

from arti.fingerprints import Fingerprint
from arti.formats import Format
from arti.internal.models import Model
from arti.internal.type_hints import get_class_type_vars, lenient_issubclass
from arti.internal.utils import frozendict
from arti.partitions import CompositeKey, CompositeKeyTypes
from arti.storage._internal import InputFingerprints as InputFingerprints
from arti.storage._internal import partial_format
from arti.types import Type

_StoragePartition = TypeVar("_StoragePartition", bound="StoragePartition")


class StoragePartition(Model):
    keys: CompositeKey
    fingerprint: Fingerprint = Fingerprint.empty()

    def with_fingerprint(self: _StoragePartition) -> _StoragePartition:
        return self.copy(update={"fingerprint": self.compute_fingerprint()})

    @abc.abstractmethod
    def compute_fingerprint(self) -> Fingerprint:
        raise NotImplementedError("{type(self).__name__}.compute_fingerprint is not implemented!")


StoragePartitions = tuple[StoragePartition, ...]  # type: ignore

_Storage = TypeVar("_Storage", bound="Storage[Any]")


class Storage(Model, Generic[_StoragePartition]):
    """Storage is a data reference identifying 1 or more partitions of data.

    Storage fields should have defaults set with placeholders for tags and partition
    keys. This allows automatic injection of the tags and partition keys for simple
    cases.
    """

    _abstract_ = True

    # These separators are used in the default resolve_* helpers to format metadata into
    # the storage fields.
    #
    # The defaults are tailored for "path"-like fields.
    key_value_sep: ClassVar[str] = "="
    partition_name_component_sep: ClassVar[str] = "_"
    segment_sep: ClassVar[str] = os.sep

    storage_partition_type: ClassVar[type[_StoragePartition]]

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if cls._abstract_:
            return
        cls.storage_partition_type = get_class_type_vars(cls)[0]
        expected_field_types = {
            name: info.outer_type_
            for name, info in cls.storage_partition_type.__fields__.items()
            if name not in StoragePartition.__fields__
        }
        fields = {
            name: info.outer_type_
            for name, info in cls.__fields__.items()
            if name not in Storage.__fields__
        }
        if fields != expected_field_types:
            raise TypeError(
                f"{cls.__name__} fields must match {cls.storage_partition_type.__name__} ({expected_field_types}), got: {fields}"
            )

    @abc.abstractmethod
    def discover_partitions(
        self,
        key_types: CompositeKeyTypes,
        input_fingerprints: InputFingerprints = InputFingerprints(),
    ) -> tuple[_StoragePartition, ...]:
        raise NotImplementedError()

    def generate_partition(
        self,
        keys: CompositeKey,
        input_fingerprint: Optional[Fingerprint],
        with_fingerprint: bool = True,
    ) -> _StoragePartition:
        kwargs = dict[Any, Any](keys)
        if input_fingerprint is not None:
            kwargs["input_fingerprint"] = str(input_fingerprint.key)
        partition = self.storage_partition_type(
            keys=keys,
            **{
                name: (
                    original.format(**kwargs)
                    if lenient_issubclass(type(original := getattr(self, name)), str)
                    else original
                )
                for name in self.__fields__
                if name in self.storage_partition_type.__fields__
            },
        )
        if with_fingerprint:
            partition = partition.with_fingerprint()
        return partition

    def _resolve_field(self, spec: str, placeholder_values: dict[str, str]) -> str:
        for placeholder, value in placeholder_values.items():
            if not value:
                # Strip placeholder *and* any trailing self.segment_sep
                trim = "{" + placeholder + "}"
                if f"{trim}{self.segment_sep}" in spec:
                    trim = f"{trim}{self.segment_sep}"
                spec = spec.replace(trim, "")
        return partial_format(spec, **placeholder_values)

    def resolve(self: _Storage, **placeholder_values: str) -> _Storage:
        return self.copy(
            update={
                name: updated_value
                for name in self.__fields__
                if lenient_issubclass(type(original := getattr(self, name)), str)
                # Avoid "setting" the value if not updated to reduce pydantic repr verbosity (which
                # only shows "set" fields by default).
                and (updated_value := self._resolve_field(original, placeholder_values)) != original
            }
        )

    def resolve_extension(self: _Storage, extension: Optional[str]) -> _Storage:
        if extension is None:
            return self
        return self.resolve(extension=extension)

    def resolve_graph_name(self: _Storage, graph_name: str) -> _Storage:
        return self.resolve(graph_name=graph_name)

    def resolve_names(self: _Storage, names: tuple[str, ...]) -> _Storage:
        return self.resolve(names=self.segment_sep.join(names), name=names[-1])

    def resolve_partition_key_spec(self: _Storage, key_types: CompositeKeyTypes) -> _Storage:
        key_component_specs = {
            f"{name}{self.partition_name_component_sep}{key_component}": (
                "{" + f"{name}.{key_component}" + "}"
            )
            for name, pk in key_types.items()
            for key_component in pk.default_key_components
        }
        return self.resolve(
            partition_key_spec=self.segment_sep.join(
                f"{name}{self.key_value_sep}{spec}" for name, spec in key_component_specs.items()
            )
        )

    def resolve_path_tags(self: _Storage, path_tags: frozendict[str, str]) -> _Storage:
        return self.resolve(
            path_tags=self.segment_sep.join(
                f"{tag}{self.key_value_sep}{value}" for tag, value in path_tags.items()
            )
        )

    def supports(self: _Storage, type_: Type, format: Format) -> None:
        # TODO: Ensure the storage supports all of the specified types and partitioning on the
        # specified field(s).
        pass



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

import abc

from arti.graphs import Graph
from arti.internal.models import Model


class Executor(Model):
    @abc.abstractmethod
    def build(self, graph: Graph) -> None:
        pass



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from collections.abc import Iterable, Iterator, Mapping
from operator import attrgetter
from typing import Any, ClassVar, Literal, Optional

from pydantic import PrivateAttr, validator
from pydantic.fields import ModelField

from arti.internal.models import Model
from arti.internal.type_hints import lenient_issubclass
from arti.internal.utils import class_name, classproperty, frozendict, register


class Type(Model):
    """Type represents a data type."""

    _abstract_ = True
    description: Optional[str]
    nullable: bool = False


class _NamedMixin(Model):
    name: str = "anon"

    @classmethod
    def _pydantic_type_system_post_field_conversion_hook_(
        cls, type_: Type, *, name: str, required: bool
    ) -> Type:
        type_ = super()._pydantic_type_system_post_field_conversion_hook_(
            type_, name=name, required=required
        )
        if "name" not in type_.__fields_set__:
            type_ = type_.copy(update={"name": name})
        return type_


########################
# Core Artigraph Types #
########################


class _Numeric(Type):
    pass


class _Float(_Numeric):
    pass


class _Int(_Numeric):
    pass


class Binary(Type):
    byte_size: Optional[int]


class Boolean(Type):
    pass


class Date(Type):
    pass


class Enum(Type, _NamedMixin):
    type: Type
    items: frozenset[Any]

    @validator("items", pre=True)
    @classmethod
    def _cast_values(cls, items: Any) -> Any:
        if isinstance(items, Iterable) and not isinstance(items, Mapping):
            return frozenset(items)
        return items

    @validator("items")
    @classmethod
    def _validate_values(cls, items: frozenset[Any], values: dict[str, Any]) -> frozenset[Any]:
        from arti.types.python import python_type_system

        if len(items) == 0:
            raise ValueError("cannot be empty.")
        # `type` will be missing if it doesn't pass validation.
        if (arti_type := values.get("type")) is None:
            return items
        py_type = python_type_system.to_system(arti_type, hints={})
        mismatched_items = [item for item in items if not lenient_issubclass(type(item), py_type)]
        if mismatched_items:
            raise ValueError(f"incompatible {arti_type} ({py_type}) item(s): {mismatched_items}")
        return items


class Float16(_Float):
    pass


class Float32(_Float):
    pass


class Float64(_Float):
    pass


class Int8(_Int):
    pass


class Int16(_Int):
    pass


class Int32(_Int):
    pass


class Int64(_Int):
    pass


class List(Type):
    element: Type
    partition_by: tuple[str, ...] = ()
    cluster_by: tuple[str, ...] = ()

    @validator("partition_by", "cluster_by")
    @classmethod
    def _validate_field_ref(
        cls, references: tuple[str, ...], values: dict[str, Any], field: ModelField
    ) -> tuple[str, ...]:
        if (element := values.get("element")) is None:
            return references
        if references and not isinstance(element, Struct):
            raise ValueError("requires element to be a Struct")
        known, requested = set(element.fields), set(references)
        if unknown := requested - known:
            raise ValueError(f"unknown field(s): {unknown}")
        return references

    @validator("cluster_by")
    @classmethod
    def _validate_cluster_by(
        cls, cluster_by: tuple[str, ...], values: dict[str, Any]
    ) -> tuple[str, ...]:
        if (partition_by := values.get("partition_by")) is None:
            return cluster_by
        if overlapping := set(cluster_by) & set(partition_by):
            raise ValueError(f"clustering fields overlap with partition fields: {overlapping}")
        return cluster_by

    @classproperty
    @classmethod
    def is_partitioned(cls) -> bool:
        return bool(cls.partition_fields)

    @property
    def partition_fields(self) -> frozendict[str, Type]:
        if not isinstance(self.element, Struct):
            return frozendict()
        return frozendict({name: self.element.fields[name] for name in self.partition_by})


class Map(Type):
    key: Type
    value: Type


class Null(Type):
    pass


class Set(Type):
    element: Type


class String(Type):
    pass


class Struct(Type, _NamedMixin):
    fields: frozendict[str, Type]


class Timestamp(Type):
    """UTC timestamp with configurable precision."""

    precision: Literal["second", "millisecond", "microsecond", "nanosecond"]


class UInt8(_Int):
    pass


class UInt16(_Int):
    pass


class UInt32(_Int):
    pass


class UInt64(_Int):
    pass


##############################
# Type conversion interfaces #
##############################


class TypeAdapter:
    """TypeAdapter maps between Artigraph types and a foreign type system."""

    key: ClassVar[str] = class_name()

    artigraph: ClassVar[type[Type]]  # The internal Artigraph Type
    system: ClassVar[Any]  # The external system's type

    priority: ClassVar[int] = 0  # Set the priority of this mapping. Higher is better.

    @classmethod
    def matches_artigraph(cls, type_: Type, *, hints: dict[str, Any]) -> bool:
        return isinstance(type_, cls.artigraph)

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> Type:
        raise NotImplementedError()

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        raise NotImplementedError()

    @classmethod
    def to_system(cls, type_: Type, *, hints: dict[str, Any]) -> Any:
        raise NotImplementedError()


# _ScalarClassTypeAdapter can be used for scalars defined as python types (eg: int or str for the
# python TypeSystem).
class _ScalarClassTypeAdapter(TypeAdapter):
    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> Type:
        return cls.artigraph()

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        return lenient_issubclass(type_, cls.system)

    @classmethod
    def to_system(cls, type_: Type, *, hints: dict[str, Any]) -> Any:
        return cls.system

    @classmethod
    def generate(
        cls,
        *,
        artigraph: type[Type],
        system: Any,
        priority: int = 0,
        type_system: TypeSystem,
        name: Optional[str] = None,
    ) -> type[TypeAdapter]:
        """Generate a _ScalarClassTypeAdapter subclass for the scalar system type."""
        name = name or f"{type_system.key}{artigraph.__name__}"
        return type_system.register_adapter(
            type(
                name,
                (cls,),
                {"artigraph": artigraph, "system": system, "priority": priority},
            )
        )


class TypeSystem(Model):
    key: str

    _adapter_by_key: dict[str, type[TypeAdapter]] = PrivateAttr(default_factory=dict)

    def register_adapter(self, adapter: type[TypeAdapter]) -> type[TypeAdapter]:
        return register(self._adapter_by_key, adapter.key, adapter)

    @property
    def _priority_sorted_adapters(self) -> Iterator[type[TypeAdapter]]:
        return reversed(sorted(self._adapter_by_key.values(), key=attrgetter("priority")))

    def to_artigraph(self, type_: Any, *, hints: dict[str, Any]) -> Type:
        for adapter in self._priority_sorted_adapters:
            if adapter.matches_system(type_, hints=hints):
                return adapter.to_artigraph(type_, hints=hints)
        raise NotImplementedError(f"No {self} adapter for system type: {type_}.")

    def to_system(self, type_: Type, *, hints: dict[str, Any]) -> Any:
        for adapter in self._priority_sorted_adapters:
            if adapter.matches_artigraph(type_, hints=hints):
                return adapter.to_system(type_, hints=hints)
        raise NotImplementedError(f"No {self} adapter for Artigraph type: {type_}.")

    def check_similarity(self, *, arti: Type, python_type: type) -> None:
        system_type = self.to_system(arti, hints={})
        if not lenient_issubclass(system_type, python_type):
            raise ValueError(f"{python_type} cannot be used to represent {arti}")



import importlib.metadata
from collections.abc import Iterator
from contextlib import contextmanager

version = importlib.metadata.version("arti")


@contextmanager
def wrap_exc(error_type: type[Exception], *, prefix: str) -> Iterator[None]:
    """Wrap exceptions of `error_type` and add a message prefix.

    `error_type` must be initializable with a single string message argument.

    NOTE: When used inside a generator, any exceptions raised by the *caller of the generator* will **not** be wrapped.
    """
    try:
        yield
    except error_type as e:
        msg = str(e)
        if getattr(e, "wrapped", False) and e.__cause__ is not None:
            src = e.__cause__  # Shorten exception chains to the root and last wrapped only
        else:
            msg = f" - {msg}"
            src = e
        error = error_type(f"{prefix}{msg}")
        error.wrapped = True  # type: ignore
        raise error from src



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from collections import defaultdict
from collections.abc import Callable
from functools import cached_property, wraps
from graphlib import TopologicalSorter
from types import TracebackType
from typing import TYPE_CHECKING, Any, Literal, Optional, TypeVar, Union, cast

from pydantic import Field, PrivateAttr

import arti
from arti import io
from arti.artifacts import Artifact
from arti.backends import Backend
from arti.backends.memory import MemoryBackend
from arti.fingerprints import Fingerprint
from arti.internal.models import Model
from arti.internal.utils import TypedBox, frozendict
from arti.partitions import CompositeKey
from arti.producers import Producer
from arti.storage import StoragePartition
from arti.views import View

if TYPE_CHECKING:
    from arti.executors import Executor
else:
    from arti.internal.patches import patch_TopologicalSorter_class_getitem

    patch_TopologicalSorter_class_getitem()

# TODO: Add GraphMetadata model


SEALED: Literal[True] = True
OPEN: Literal[False] = False
BOX_KWARGS = {
    status: {
        "default_box": status is OPEN,
        "frozen_box": status is SEALED,
    }
    for status in (OPEN, SEALED)
}

_Return = TypeVar("_Return")


def requires_sealed(fn: Callable[..., _Return]) -> Callable[..., _Return]:
    @wraps(fn)
    def check_if_sealed(self: Graph, *args: Any, **kwargs: Any) -> _Return:
        if self._status is not SEALED:
            raise ValueError(f"{fn.__name__} cannot be used while the Graph is still being defined")
        return fn(self, *args, **kwargs)

    return check_if_sealed


class ArtifactBox(TypedBox[str, Artifact]):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        object.__setattr__(self, "_path", ())
        super().__init__(*args, **kwargs)

    def _TypedBox__cast_value(self, item: str, artifact: Any) -> Artifact:
        artifact = super()._TypedBox__cast_value(item, artifact)  # type: ignore
        if (graph := arti.context.graph) is not None:
            artifact = artifact.copy(
                update={
                    "storage": (
                        artifact.storage.resolve_graph_name(graph.name)
                        .resolve_names(self._path + (item,))
                        .resolve_path_tags(graph.path_tags)
                    )
                }
            )
        return cast(Artifact, artifact)

    def _Box__convert_and_store(self, item: str, value: Artifact) -> None:
        if isinstance(value, dict):
            super()._Box__convert_and_store(item, value)  # pylint: disable=no-member
            # TODO: Test if this works with `Box({"some": {"nested": "thing"}})`.
            # Guessing not, may need to put in an empty dict/box first, set the path,
            # and then update it.
            object.__setattr__(self[item], "_path", self._path + (item,))
        else:
            super()._Box__convert_and_store(item, value)  # pylint: disable=no-member

    def _Box__get_default(self, item: str, attr: bool = False) -> Any:
        value = super()._Box__get_default(item, attr=attr)  # type: ignore
        if isinstance(value, ArtifactBox):
            object.__setattr__(value, "_path", self._path + (item,))
        return value


Node = Union[Artifact, Producer]
NodeDependencies = frozendict[Node, frozenset[Node]]


class Graph(Model):
    """Graph stores a web of Artifacts connected by Producers."""

    name: str
    backend: Backend = Field(default_factory=MemoryBackend)
    path_tags: frozendict[str, str] = frozendict()

    # Graph starts off sealed, but is opened within a `with Graph(...)` context
    _status: Optional[bool] = PrivateAttr(None)
    _artifacts: ArtifactBox = PrivateAttr(default_factory=lambda: ArtifactBox(**BOX_KWARGS[SEALED]))

    def __enter__(self) -> Graph:
        if arti.context.graph is not None:
            raise ValueError(f"Another graph is already in use: {arti.context.graph}")
        arti.context.graph = self
        self._toggle(OPEN)
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        arti.context.graph = None
        self._toggle(SEALED)
        # Confirm the dependencies are acyclic
        TopologicalSorter(self.dependencies).prepare()

    def _toggle(self, status: bool) -> None:
        self._status = status
        self._artifacts = ArtifactBox(self.artifacts, **BOX_KWARGS[status])

    @property
    def artifacts(self) -> ArtifactBox:
        return self._artifacts

    @requires_sealed
    def build(self, executor: Optional[Executor] = None) -> None:
        if executor is None:
            from arti.executors.local import LocalExecutor

            executor = LocalExecutor()
        # TODO: Resolve and statically set all available fingerprints. Specifically, we
        # should pin the Producer.fingerprint, which may by dynamic (eg: version is a
        # Timestamp). Unbuilt Artifact (partitions) won't be fully resolved yet.
        return executor.build(self)

    @cached_property  # type: ignore # python/mypy#1362
    @requires_sealed
    def dependencies(self) -> NodeDependencies:
        artifact_deps = {
            artifact: (
                frozenset({artifact.producer_output.producer})
                if artifact.producer_output is not None
                else frozenset()
            )
            for key, artifact in self.artifacts.walk()
        }
        producer_deps = {
            # NOTE: multi-output Producers will appear multiple times (but be deduped)
            producer_output.producer: frozenset(producer_output.producer.inputs.values())
            for artifact in artifact_deps
            if (producer_output := artifact.producer_output) is not None
        }
        return NodeDependencies(artifact_deps | producer_deps)  # type: ignore

    @cached_property  # type: ignore # python/mypy#1362
    @requires_sealed
    def producers(self) -> frozenset[Producer]:
        return frozenset(self.producer_outputs)

    @cached_property  # type: ignore # python/mypy#1362
    @requires_sealed
    def producer_outputs(self) -> frozendict[Producer, tuple[Artifact, ...]]:
        d = defaultdict[Producer, dict[int, Artifact]](dict)
        for _, artifact in self.artifacts.walk():
            if artifact.producer_output is None:
                continue
            output = artifact.producer_output
            d[output.producer][output.position] = artifact
        return frozendict(
            (producer, tuple(artifacts_by_position[i] for i in sorted(artifacts_by_position)))
            for producer, artifacts_by_position in d.items()
        )

    # TODO: io.read/write probably need a bit of sanity checking (probably somewhere
    # else), eg: type ~= view. Doing validation on the data, etc. Should some of this
    # live on the View?

    def read(
        self,
        artifact: Artifact,
        *,
        storage_partitions: Optional[tuple[StoragePartition, ...]] = None,
        annotation: Optional[Any] = None,
        view: Optional[View] = None,
    ) -> Any:
        if annotation is not None:
            if view is not None:
                raise ValueError("Only one of `annotation` or `view` may be set")
            view = View.get_class_for(annotation)()
        assert view is not None
        if storage_partitions is None:
            with self.backend.connect() as backend:
                storage_partitions = backend.read_artifact_partitions(artifact)
        return io.read(
            type=artifact.type,
            format=artifact.format,
            storage_partitions=storage_partitions,
            view=view,
        )

    def write(
        self,
        data: Any,
        *,
        artifact: Artifact,
        input_fingerprint: Optional[Fingerprint] = None,
        keys: CompositeKey = CompositeKey(),
        view: Optional[View] = None,
    ) -> StoragePartition:
        if view is None:
            view = View.get_class_for(type(data))()
        storage_partition = artifact.storage.generate_partition(
            input_fingerprint=input_fingerprint, keys=keys, with_fingerprint=False
        )
        io.write(
            data,
            type=artifact.type,
            format=artifact.format,
            storage_partition=storage_partition,
            view=view,
        )
        storage_partition = storage_partition.with_fingerprint()
        # TODO: Should we only do this in bulk? We might want the backends to
        # transparently batch requests, but that's not so friendly with the transient
        # ".connect".
        with self.backend.connect() as backend:
            backend.write_artifact_partitions(artifact, (storage_partition,))
        return cast(StoragePartition, storage_partition)



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from arti.artifacts import Statistic as Statistic  # noqa: F401

# class FieldStatistic(Statistic):
#     """ A FieldStatistic is a Statistic associated with a particular Artifact field.
#     """
#
#
# class Count(FieldStatistic):
#     type = Int64()
#
#
# class CountDistinct(FieldStatistic):
#     type = Int64()
#
#
# class MaxInt64(FieldStatistic):
#     type = Int64()
#
#
# class MinInt64(FieldStatistic):
#     type = Int64()
#
#
# class SumInt64(FieldStatistic):
#     type = Int64()



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from arti.internal.models import Model


class Annotation(Model):
    """An Annotation is a piece of human knowledge associated with an Artifact."""

    _abstract_ = True



from functools import partial
from typing import Any, Callable, ClassVar, Literal, Optional, cast

import sgqlc.types as st

import arti.types as at
from arti.types import TypeAdapter, TypeSystem, _ScalarClassTypeAdapter


def is_nullable(type_: type[st.BaseType]) -> bool:
    return not type_.__name__.endswith("!")


def is_list(type_: type[st.BaseType]) -> bool:
    return type_.__name__.startswith("[")


def get_schema(hints: dict[str, Any]) -> st.Schema:
    if (schema := hints.get(SGQLC_HINT_SCHEMA)) is not None:
        return cast(st.Schema, schema)
    # Avoid referencing (and thus polluting) the global schema.
    return st.Schema()


def get_existing(
    type_: at.Type,
    schema: st.Schema,
    equals: Callable[[at.Type, type[st.BaseType]], bool],
) -> Optional[type[st.BaseType]]:
    assert isinstance(type_, at._NamedMixin)
    if (existing := getattr(schema, type_.name, None)) is not None:
        if not equals(type_, existing):
            raise ValueError(
                f"Detected duplicate but mismatched sgqlc {existing.__name__} type when converting {type_}: {repr(existing)}"
            )
        return cast(type[st.BaseType], existing)
    return None


# sgqlc types are implicitly nullable, which is the opposite of the arti Type default. Rather than
# handle nullable in every TypeAdapter below, we'll handle in the base dispatch.
#
# This differs from the python_type_system, where not-null is the default AND there's an Optional
# type we can easily match against.
class _SgqlcTypeSystem(TypeSystem):
    def to_artigraph(self, type_: Any, *, hints: dict[str, Any]) -> at.Type:
        if not (nullable := is_nullable(type_)):
            # non_null subclasses the input type (hence, we can fetch it with .mro)
            type_ = type_.mro()[1]
        ret = super().to_artigraph(type_, hints=hints)
        if ret.nullable != nullable:
            ret = ret.copy(update={"nullable": nullable})
        return ret

    def to_system(self, type_: at.Type, *, hints: dict[str, Any]) -> Any:
        ret = super().to_system(type_.copy(update={"nullable": True}), hints=hints)
        assert is_nullable(ret)  # sgqlc default
        if not type_.nullable:
            ret = st.non_null(ret)
        return ret


sgqlc_type_system = _SgqlcTypeSystem(key="sgqlc")

SGQLC_HINT_ABSTRACT = f"{sgqlc_type_system.key}.abstract"
SGQLC_HINT_INTERFACES = f"{sgqlc_type_system.key}.interfaces"
SGQLC_HINT_SCHEMA = f"{sgqlc_type_system.key}.schema"

_generate = partial(_ScalarClassTypeAdapter.generate, type_system=sgqlc_type_system)

_generate(artigraph=at.Boolean, system=st.Boolean)
for _precision in (16, 32, 64):
    _generate(
        artigraph=getattr(at, f"Float{_precision}"),
        system=st.Float,
        priority=_precision,
    )
for _precision in (8, 16, 32, 64):
    _generate(
        artigraph=getattr(at, f"Int{_precision}"),
        system=st.Int,
        priority=_precision,
    )
# Register st.String with higher priority to avoid at.String->ID by default
_generate(artigraph=at.String, system=st.String, priority=1)
_generate(artigraph=at.String, system=st.ID, name="sgqlcID")


# NOTE: GraphQL only supports string enums (a series of "names")
@sgqlc_type_system.register_adapter
class SgqlcEnumAdapter(TypeAdapter):
    artigraph = at.Enum
    system = st.Enum

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> at.Type:
        assert issubclass(type_, cls.system)
        return cls.artigraph(
            name=type_.__name__,
            type=at.String(),
            items=type_.__choices__,
        )

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        return issubclass(type_, cls.system)

    @classmethod
    def _compare_existing(cls, type_: at.Type, existing: type[st.BaseType]) -> bool:
        return cls.matches_system(existing, hints={}) and cast(at.Enum, type_).items == set(
            existing  # type: ignore
        )

    @classmethod
    def to_system(cls, type_: at.Type, *, hints: dict[str, Any]) -> Any:
        assert isinstance(type_, cls.artigraph)
        assert isinstance(type_.type, at.String)
        schema = get_schema(hints)
        if (existing := get_existing(type_, schema, equals=cls._compare_existing)) is not None:
            return existing
        return type(
            type_.name,
            (cls.system,),
            {
                "__choices__": tuple(type_.items),
                "__schema__": schema,
            },
        )


@sgqlc_type_system.register_adapter
class SgqlcList(TypeAdapter):
    artigraph = at.List
    system = st.BaseType
    priority = int(1e9)  # May wrap other types, so must be highest priority

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> at.Type:
        assert issubclass(type_, cls.system)
        # list_of subclasses the input type (hence, we can fetch it with .mro)
        return cls.artigraph(element=sgqlc_type_system.to_artigraph(type_.mro()[1], hints=hints))

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        return issubclass(type_, cls.system) and is_list(type_)

    @classmethod
    def to_system(cls, type_: at.Type, *, hints: dict[str, Any]) -> Any:
        assert isinstance(type_, cls.artigraph)
        return st.list_of(sgqlc_type_system.to_system(type_.element, hints=hints))


class _StructAdapter(TypeAdapter):
    artigraph = at.Struct
    kind: ClassVar[Literal["interface", "type"]]

    @classmethod
    def matches_artigraph(cls, type_: at.Type, *, hints: dict[str, Any]) -> bool:
        abstract = hints.get(SGQLC_HINT_ABSTRACT, False)
        # Interfaces should be abstract, Types not.
        return isinstance(type_, cls.artigraph) and (
            abstract if cls.kind == "interface" else not abstract
        )

    @classmethod
    def to_artigraph(cls, type_: Any, *, hints: dict[str, Any]) -> at.Type:
        assert issubclass(type_, cls.system)
        return cls.artigraph(
            name=type_.__name__,
            fields={
                field.name: sgqlc_type_system.to_artigraph(field.type, hints=hints)
                for field in type_
            },
        )

    @classmethod
    def matches_system(cls, type_: Any, *, hints: dict[str, Any]) -> bool:
        return issubclass(type_, cls.system) and type_.__kind__ == cls.kind

    @classmethod
    def _compare_existing(
        cls, type_: at.Type, existing: type[st.BaseType], field_types: dict[str, type[st.BaseType]]
    ) -> bool:
        assert set(cast(at.Struct, type_).fields) == set(field_types)
        return cls.matches_system(existing, hints={}) and field_types == {
            field.name: field.type for field in cast(st.Type, existing)
        }

    @classmethod
    def to_system(cls, type_: at.Type, *, hints: dict[str, Any]) -> Any:
        assert isinstance(type_, cls.artigraph)
        schema = get_schema(hints)
        fields = {k: sgqlc_type_system.to_system(v, hints=hints) for k, v in type_.fields.items()}
        compare_existing = partial(cls._compare_existing, field_types=fields)
        if (existing := get_existing(type_, schema, equals=compare_existing)) is not None:
            return existing
        return type(
            type_.name,
            (
                cls.system,
                *hints.get(SGQLC_HINT_INTERFACES, ()),
            ),
            {
                "__schema__": schema,
                **fields,
            },
        )


@sgqlc_type_system.register_adapter
class SgqlcInterfaceAdapter(_StructAdapter):
    system = st.Interface
    kind: ClassVar[Literal["interface"]] = "interface"


@sgqlc_type_system.register_adapter
class SgqlcTypeAdapter(_StructAdapter):
    system = st.Type
    kind: ClassVar[Literal["type"]] = "type"



from __future__ import annotations

import functools
import importlib
import inspect
import os.path
import pkgutil
import threading
from collections.abc import Callable, Generator, Iterator, MutableMapping
from contextlib import contextmanager
from tempfile import TemporaryDirectory
from types import GenericAlias, ModuleType
from typing import IO, Any, ClassVar, Generic, Optional, TypeVar, Union, cast, overload

from box import Box
from frozendict.core import frozendict as _frozendict
from multimethod import multidispatch
from setuptools import find_namespace_packages
from toolz import curry

from arti.internal.type_hints import lenient_issubclass, tidy_signature

_K = TypeVar("_K")
_V = TypeVar("_V")


class ClassName:
    def __get__(self, obj: Any, type_: type[Any]) -> str:
        return type_.__name__


class_name = cast(Callable[[], str], ClassName)


PropReturn = TypeVar("PropReturn")


class classproperty(Generic[PropReturn]):
    """Access a @classmethod like a @property.

    Can be stacked above @classmethod (to satisfy pylint, mypy, etc).
    """

    def __init__(self, f: Callable[..., PropReturn]) -> None:
        self.f = f
        if isinstance(self.f, classmethod):
            self.f = lambda type_: f.__get__(None, type_)()

    def __get__(self, obj: Any, type_: Any) -> PropReturn:
        return self.f(type_)


RETURN = TypeVar("RETURN")
REGISTERED = TypeVar("REGISTERED", bound=Callable[..., Any])


# This may be less useful once mypy supports ParamSpecs - after that, we might be able to define
# multidispatch with a ParamSpec and have mypy check the handlers' arguments are covariant.
@curry
class dispatch(multidispatch[RETURN]):
    """Multiple dispatch for a set of functions based on parameter type.

    Usage is similar to `@functools.singledispatch`. The original definition defines the "spec" that
    subsequent handlers must follow, namely the name and (base)class of parameters.
    """

    # multimethod/multidispatch __new__/__init__ work a bit specially - __new__ doesn't
    # call __init__, but instead

    # Modified from multidispatch to allow our extra kwargs
    def __new__(
        cls, func: Callable[..., RETURN], once_before: Callable[[], None] = lambda: None
    ) -> "dispatch[RETURN]":
        return functools.update_wrapper(dict.__new__(cls), func)

    def __init__(
        self, func: Callable[..., RETURN], once_before: Callable[[], None] = lambda: None
    ) -> None:
        super().__init__(func)
        self.clean_signature = tidy_signature(func, self.signature)
        if self.clean_signature.return_annotation not in (Any, type(None)):
            raise NotImplementedError("Return type checking is not implemented yet!")
        self.once_before = once_before
        self.once_before_finished = False

    def __call__(self, *args: Any, **kwargs: Any) -> RETURN:
        if not self.once_before_finished:
            self.once_before()
            self.once_before_finished = True
        return super().__call__(*args, **kwargs)

    @overload
    def register(self, __func: REGISTERED) -> REGISTERED:
        ...

    @overload
    def register(self, *args: type) -> Callable[[REGISTERED], REGISTERED]:
        ...

    def register(self, *args: Any) -> Callable[[REGISTERED], REGISTERED]:
        if len(args) == 1 and hasattr(args[0], "__annotations__"):
            func = args[0]
            sig = tidy_signature(func, inspect.signature(func))
            spec = self.clean_signature
            if set(sig.parameters) != set(spec.parameters):
                raise TypeError(
                    f"Expected `{func.__name__}` to have {sorted(set(spec.parameters))} parameters, got {sorted(set(sig.parameters))}"
                )
            for name in sig.parameters:
                sig_param, spec_param = sig.parameters[name], spec.parameters[name]
                if sig_param.kind != spec_param.kind:
                    raise TypeError(
                        f"Expected the `{func.__name__}.{name}` parameter to be {spec_param.kind}, got {sig_param.kind}"
                    )
                if sig_param.annotation is not Any and not lenient_issubclass(
                    sig_param.annotation, spec_param.annotation
                ):
                    raise TypeError(
                        f"Expected the `{func.__name__}.{name}` parameter to be a subclass of {spec_param.annotation}, got {sig_param.annotation}"
                    )
        return super().register(*args)  # type: ignore


# frozendict is useful for models to preserve hashability (eg: key in a dict). Unfortunately, there
# is 1 small pydantic issue[1] and an upstream issue with types.GenericAlias + deepcopy[2,3]. We can
# work around the blocking deepcopy issue by replacing types.GenericAlias with typing.Generic (which
# returns a typing._GenericAlias). The pydantic bug (which returns a dict instead of a frozendict)
# is worked around in arti.internal.patches.patch_pydantic_ModelField__type_analysis.
#
# 1: https://github.com/samuelcolvin/pydantic/pull/3138
# 2: https://github.com/Marco-Sulla/python-frozendict/issues/29
# 3: https://bugs.python.org/issue45167
class frozendict(Generic[_K, _V], _frozendict[_K, _V]):
    pass


def import_submodules(
    path: list[str],  # module.__path__ is a list[str]
    name: str,
    *,
    lock: threading.Lock = threading.Lock(),
) -> dict[str, ModuleType]:
    """Recursively import submodules.

    This can be useful with registry patterns to automatically discover and import submodules
    defining additional implementations.

    `path` and `name` are usually provided from an existing module's `__path__` and `__name__`.

    This function is thread-safe and supports namespace modules.

    NOTE: This inherently triggers eager imports, which has performance impacts and may cause import
    cycles. To reduce these issues, avoid calling during module definition.
    """
    # pkgutil.iter_modules is not recursive and pkgutil.walk_packages does not handle namespace
    # packages... however we can leverage setuptools.find_namespace_packages, which is build for
    # exactly this.
    path_names = {p: name for p in path}
    path_names.update(
        {
            os.sep.join([path, *name.split(".")]): f"{root_name}.{name}"
            for path, root_name in path_names.items()
            for name in find_namespace_packages(path)
        }
    )
    with lock:
        return {
            name: importlib.import_module(name)
            for path, name in path_names.items()
            for _, name, _ in pkgutil.iter_modules([path], prefix=f"{name}.")
        }


_int_sub = TypeVar("_int_sub", bound="_int")


class _int(int):
    def __repr__(self) -> str:
        return f"{qname(self)}({int(self)})"

    def __str__(self) -> str:
        return str(int(self))

    # Stock magics.
    #
    # Using "self: TypeVar" so mypy will detect the returned subclass (rather than _int).

    def __add__(self: _int_sub, x: int) -> _int_sub:
        return type(self)(super().__add__(x))

    def __and__(self: _int_sub, n: int) -> _int_sub:
        return type(self)(super().__and__(n))

    def __ceil__(self: _int_sub) -> _int_sub:
        return type(self)(super().__ceil__())

    def __floor__(self: _int_sub) -> _int_sub:
        return type(self)(super().__floor__())

    def __floordiv__(self: _int_sub, x: int) -> _int_sub:
        return type(self)(super().__floordiv__(x))

    def __invert__(self: _int_sub) -> _int_sub:
        return type(self)(super().__invert__())

    def __lshift__(self: _int_sub, n: int) -> _int_sub:
        return type(self)(super().__lshift__(n))

    def __mod__(self: _int_sub, x: int) -> _int_sub:
        return type(self)(super().__mod__(x))

    def __mul__(self: _int_sub, x: int) -> _int_sub:
        return type(self)(super().__mul__(x))

    def __neg__(self: _int_sub) -> _int_sub:
        return type(self)(super().__neg__())

    def __or__(self: _int_sub, n: int) -> _int_sub:
        return type(self)(super().__or__(n))

    def __pos__(self: _int_sub) -> _int_sub:
        return type(self)(super().__pos__())

    def __radd__(self: _int_sub, x: int) -> _int_sub:
        return type(self)(super().__radd__(x))

    def __rand__(self: _int_sub, n: int) -> _int_sub:
        return type(self)(super().__rand__(n))

    def __rfloordiv__(self: _int_sub, x: int) -> _int_sub:
        return type(self)(super().__rfloordiv__(x))

    def __rlshift__(self: _int_sub, n: int) -> _int_sub:
        return type(self)(super().__rlshift__(n))

    def __rmod__(self: _int_sub, x: int) -> _int_sub:
        return type(self)(super().__rmod__(x))

    def __rmul__(self: _int_sub, x: int) -> _int_sub:
        return type(self)(super().__rmul__(x))

    def __ror__(self: _int_sub, n: int) -> _int_sub:
        return type(self)(super().__ror__(n))

    def __round__(self: _int_sub, ndigits: Optional[int] = 0) -> _int_sub:
        return type(self)(super().__round__(ndigits))

    def __rrshift__(self: _int_sub, n: int) -> _int_sub:
        return type(self)(super().__rrshift__(n))

    def __rshift__(self: _int_sub, n: int) -> _int_sub:
        return type(self)(super().__rshift__(n))

    def __rsub__(self: _int_sub, x: int) -> _int_sub:
        return type(self)(super().__rsub__(x))

    def __rxor__(self: _int_sub, n: int) -> _int_sub:
        return type(self)(super().__rxor__(n))

    def __sub__(self: _int_sub, x: int) -> _int_sub:
        return type(self)(super().__sub__(x))

    def __trunc__(self: _int_sub) -> _int_sub:
        return type(self)(super().__trunc__())

    def __xor__(self: _int_sub, n: int) -> _int_sub:
        return type(self)(super().__xor__(n))


class int64(_int):
    _min, _max = -(2 ** 63), (2 ** 63) - 1

    def __new__(cls, i: Union[int, int64, uint64]) -> int64:
        if i > cls._max:
            if isinstance(i, uint64):
                i = int(i) - uint64._max - 1
            else:
                raise ValueError(f"{i} is too large for int64. Hint: cast to uint64 first.")
        if i < cls._min:
            raise ValueError(f"{i} is too small for int64.")
        return super().__new__(cls, i)


class uint64(_int):
    _min, _max = 0, (2 ** 64) - 1

    def __new__(cls, i: Union[int, int64, uint64]) -> uint64:
        if i > cls._max:
            raise ValueError(f"{i} is too large for uint64.")
        if i < cls._min:
            if isinstance(i, int64):
                i = int(i) + cls._max + 1
            else:
                raise ValueError(f"{i} is negative. Hint: cast to int64 first.")
        return super().__new__(cls, i)


@contextmanager
def named_temporary_file(mode: str = "w+b") -> Generator[IO[Any], None, None]:
    """Minimal alternative to tempfile.NamedTemporaryFile that can be re-opened on Windows."""
    with TemporaryDirectory() as d:
        with open(os.path.join(d, "contents"), mode=mode) as f:
            yield f


def ordinal(n: int) -> str:
    """Convert an integer into its ordinal representation."""
    n = int(n)
    suffix = ["th", "st", "nd", "rd", "th"][min(n % 10, 4)]
    if 11 <= (n % 100) <= 13:
        suffix = "th"
    return str(n) + suffix


def register(
    registry: dict[_K, _V],
    key: _K,
    value: _V,
    get_priority: Optional[Callable[[_V], int]] = None,
) -> _V:
    if key in registry:
        existing = registry[key]
        if get_priority is None:
            raise ValueError(f"{key} is already registered with: {existing}!")
        existing_priority, new_priority = get_priority(existing), get_priority(value)
        if existing_priority > new_priority:
            return value
        if existing_priority == new_priority:
            raise ValueError(
                f"{key} with matching priority ({existing_priority}) is already registered with: {existing}!"
            )
    registry[key] = value
    return value


def qname(val: Union[object, type]) -> str:
    if isinstance(val, type):
        return val.__qualname__
    return type(val).__qualname__


class TypedBox(Box, MutableMapping[_K, Union[_V, MutableMapping[_K, _V]]]):
    """TypedBox holds a collection of typed values.

    Subclasses must set the __target_type__ to a base class for the contained values.
    """

    __target_type__: ClassVar[type[_V]]

    @classmethod
    def __class_getitem__(cls, item: tuple[type[_K], type[_V]]) -> GenericAlias:
        if not isinstance(item, tuple) or len(item) != 2:
            raise TypeError(f"{cls.__name__} expects a key and value type")
        key_type, value_type = item
        if key_type is not str:
            raise TypeError(f"{cls.__name__} key must be `str`")
        return GenericAlias(
            type(
                cls.__name__,
                (cls,),
                {
                    "__module__": __name__,
                    "__target_type__": value_type,
                },
            ),
            item,
        )

    def __setattr__(self, key: str, value: Any) -> None:
        # GenericAlias sets __orig_class__ after __init__, so preempt Box from storing that (or
        # erroring if frozen).
        if key == "__orig_class__":
            return object.__setattr__(self, key, value)
        super().__setattr__(key, value)

    def __cast_value(self, item: str, value: Any) -> _V:
        if isinstance(value, self.__target_type__):
            return value
        tgt_name = self.__target_type__.__name__
        if hasattr(self.__target_type__, "cast"):
            casted = cast(Any, self.__target_type__).cast(value)
            if isinstance(casted, self.__target_type__):
                return casted
            raise TypeError(
                f"Expected {tgt_name}.cast({value}) to return an instance of {tgt_name}, got: {casted}"
            )
        raise TypeError(f"Expected an instance of {tgt_name}, got: {value}")

    # NOTE: Box uses name mangling (double __) to prevent conflicts with contained values.
    def _Box__convert_and_store(self, item: str, value: _V) -> None:
        if isinstance(value, dict):
            super()._Box__convert_and_store(item, value)  # pylint: disable=no-member
        elif item in self:
            raise ValueError(f"{item} is already set!")
        else:
            super()._Box__convert_and_store(item, self.__cast_value(item, value))

    def walk(self, root: tuple[_K, ...] = ()) -> Iterator[tuple[tuple[_K, ...], _V]]:
        for k, v in self.items():
            subroot = root + (k,)
            if isinstance(v, TypedBox):
                yield from v.walk(root=subroot)
            else:
                yield subroot, v  # type: ignore



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from functools import partial
from typing import Any

from arti.formats import Format
from arti.internal.utils import dispatch, import_submodules
from arti.storage import StoragePartition
from arti.types import Type
from arti.views import View

# TODO write/read partitioned data, column subset

# TODO: Auto import io/ submodules (like View) - perhaps add decorator on read/write to
# lazily trigger the import on first use.

# import_submodules(__path__, __name__)

_import_submodules = partial(import_submodules, __path__, __name__)


@dispatch(once_before=_import_submodules)  # type: ignore
def read(
    type: Type, format: Format, storage_partitions: tuple[StoragePartition, ...], view: View
) -> Any:
    raise NotImplementedError(
        f"Read into {view} view from {format} format in {storage_partitions} storage not implemented."
    )


@dispatch(once_before=_import_submodules)  # type: ignore
def write(
    data: Any, type: Type, format: Format, storage_partition: StoragePartition, view: View
) -> None:
    raise NotImplementedError(
        f"Write from {view} view into {format} format in {storage_partition} storage not implemented."
    )



from typing import no_type_check


def patch_TopologicalSorter_class_getitem() -> None:
    """Patch adding TopologicalSorter.__class_getitem__ to support subscription.

    TopologicalSorter is considered Generic in typeshed (hence mypy expects a type arg),
    but is not at runtime.

    Pending: https://github.com/python/cpython/pull/28714
    """
    from graphlib import TopologicalSorter
    from types import GenericAlias

    if not hasattr(TopologicalSorter, "__class_getitem__"):
        TopologicalSorter.__class_getitem__ = classmethod(GenericAlias)  # type: ignore


@no_type_check
def patch_pydantic_ModelField__type_analysis() -> None:  # noqa # pragma: no cover
    """Patch to work around a bug causing dict subclasses to be converted to stock dicts[1].

    A PR to fix has been submitted[2] but not yet released - this patch pulls from that.

    This is required to use arti.internal.utils.frozendict as a Model attribute.

    1: https://github.com/samuelcolvin/pydantic/issues/3122
    2: https://github.com/samuelcolvin/pydantic/pull/3138
    """

    from pydantic.version import VERSION

    if tuple(int(i) for i in VERSION.split(".")) >= (1, 9):
        return

    from collections.abc import Hashable as CollectionsHashable
    from collections.abc import Iterable as CollectionsIterable
    from typing import (
        Any,
        Counter,
        DefaultDict,
        Deque,
        Dict,
        FrozenSet,
        Iterable,
        List,
        Mapping,
        Pattern,
        Sequence,
        Set,
        Tuple,
        Type,
        TypeVar,
        Union,
    )

    from pydantic.class_validators import Validator
    from pydantic.fields import (
        SHAPE_DEFAULTDICT,
        SHAPE_DEQUE,
        SHAPE_DICT,
        SHAPE_FROZENSET,
        SHAPE_GENERIC,
        SHAPE_ITERABLE,
        SHAPE_LIST,
        SHAPE_MAPPING,
        SHAPE_NAME_LOOKUP,
        SHAPE_SEQUENCE,
        SHAPE_SET,
        SHAPE_TUPLE,
        SHAPE_TUPLE_ELLIPSIS,
        ModelField,
        Undefined,
    )
    from pydantic.types import Json, JsonWrapper
    from pydantic.typing import (
        Callable,
        NoneType,
        display_as_type,
        get_args,
        get_origin,
        is_literal_type,
        is_new_type,
        is_typeddict,
        new_type_supertype,
    )
    from pydantic.utils import lenient_issubclass
    from typing_extensions import Annotated

    #################
    # SHAPE_COUNTER #
    #################

    SHAPE_COUNTER = 14
    SHAPE_NAME_LOOKUP[SHAPE_COUNTER] = "Counter[{}]"

    ###################
    # is_union_origin #
    ###################

    def is_union_origin(tp: Type[Any]) -> bool:
        return tp is Union

    ##################
    # _type_analysis #
    ##################

    def _type_analysis(self) -> None:  # noqa: C901 (ignore complexity)
        # typing interface is horrible, we have to do some ugly checks
        if lenient_issubclass(self.type_, JsonWrapper):
            self.type_ = self.type_.inner_type
            self.parse_json = True
        elif lenient_issubclass(self.type_, Json):
            self.type_ = Any
            self.parse_json = True
        elif isinstance(self.type_, TypeVar):
            if self.type_.__bound__:
                self.type_ = self.type_.__bound__
            elif self.type_.__constraints__:
                self.type_ = Union[self.type_.__constraints__]
            else:
                self.type_ = Any
        elif is_new_type(self.type_):
            self.type_ = new_type_supertype(self.type_)

        if self.type_ is Any or self.type_ is object:
            if self.required is Undefined:
                self.required = False
            self.allow_none = True
            return
        elif self.type_ is Pattern:
            # python 3.7 only, Pattern is a typing object but without sub fields
            return
        elif is_literal_type(self.type_):
            return
        elif is_typeddict(self.type_):
            return

        origin = get_origin(self.type_)
        # add extra check for `collections.abc.Hashable` for python 3.10+ where origin is not `None`
        if origin is None or origin is CollectionsHashable:
            # field is not "typing" object eg. Union, Dict, List etc.
            # allow None for virtual superclasses of NoneType, e.g. Hashable
            if isinstance(self.type_, type) and isinstance(None, self.type_):
                self.allow_none = True
            return
        elif origin is Annotated:
            self.type_ = get_args(self.type_)[0]
            self._type_analysis()
            return
        elif origin is Callable:
            return
        elif is_union_origin(origin):
            types_ = []
            for type_ in get_args(self.type_):
                if type_ is NoneType:
                    if self.required is Undefined:
                        self.required = False
                    self.allow_none = True
                    continue
                types_.append(type_)

            if len(types_) == 1:
                # Optional[]
                self.type_ = types_[0]
                # this is the one case where the "outer type" isn't just the original type
                self.outer_type_ = self.type_
                # re-run to correctly interpret the new self.type_
                self._type_analysis()
            else:
                self.sub_fields = [
                    self._create_sub_type(t, f"{self.name}_{display_as_type(t)}") for t in types_
                ]
            return
        elif issubclass(origin, Tuple):
            # origin == Tuple without item type
            args = get_args(self.type_)
            if not args:  # plain tuple
                self.type_ = Any
                self.shape = SHAPE_TUPLE_ELLIPSIS
            elif len(args) == 2 and args[1] is Ellipsis:  # e.g. Tuple[int, ...]
                self.type_ = args[0]
                self.shape = SHAPE_TUPLE_ELLIPSIS
                self.sub_fields = [self._create_sub_type(args[0], f"{self.name}_0")]
            elif args == ((),):  # Tuple[()] means empty tuple
                self.shape = SHAPE_TUPLE
                self.type_ = Any
                self.sub_fields = []
            else:
                self.shape = SHAPE_TUPLE
                self.sub_fields = [
                    self._create_sub_type(t, f"{self.name}_{i}") for i, t in enumerate(args)
                ]
            return
        elif issubclass(origin, List):
            # Create self validators
            get_validators = getattr(self.type_, "__get_validators__", None)
            if get_validators:
                self.class_validators.update(
                    {
                        f"list_{i}": Validator(validator, pre=True)
                        for i, validator in enumerate(get_validators())
                    }
                )

            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_LIST
        elif issubclass(origin, Set):
            # Create self validators
            get_validators = getattr(self.type_, "__get_validators__", None)
            if get_validators:
                self.class_validators.update(
                    {
                        f"set_{i}": Validator(validator, pre=True)
                        for i, validator in enumerate(get_validators())
                    }
                )

            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_SET
        elif issubclass(origin, FrozenSet):
            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_FROZENSET
        elif issubclass(origin, Deque):
            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_DEQUE
        elif issubclass(origin, Sequence):
            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_SEQUENCE
        # priority to most common mapping: dict
        elif origin is dict or origin is Dict:
            self.key_field = self._create_sub_type(
                get_args(self.type_)[0], "key_" + self.name, for_keys=True
            )
            self.type_ = get_args(self.type_)[1]
            self.shape = SHAPE_DICT
        elif issubclass(origin, DefaultDict):
            self.key_field = self._create_sub_type(
                get_args(self.type_)[0], "key_" + self.name, for_keys=True
            )
            self.type_ = get_args(self.type_)[1]
            self.shape = SHAPE_DEFAULTDICT
        elif issubclass(origin, Counter):
            self.key_field = self._create_sub_type(
                get_args(self.type_)[0], "key_" + self.name, for_keys=True
            )
            self.type_ = int
            self.shape = SHAPE_COUNTER
        elif issubclass(origin, Mapping):
            self.key_field = self._create_sub_type(
                get_args(self.type_)[0], "key_" + self.name, for_keys=True
            )
            self.type_ = get_args(self.type_)[1]
            self.shape = SHAPE_MAPPING
        # Equality check as almost everything inherits form Iterable, including str
        # check for Iterable and CollectionsIterable, as it could receive one even when declared with the other
        elif origin in {Iterable, CollectionsIterable}:
            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_ITERABLE
            self.sub_fields = [self._create_sub_type(self.type_, f"{self.name}_type")]
        elif issubclass(origin, Type):
            return
        elif hasattr(origin, "__get_validators__") or self.model_config.arbitrary_types_allowed:
            # Is a Pydantic-compatible generic that handles itself
            # or we have arbitrary_types_allowed = True
            self.shape = SHAPE_GENERIC
            self.sub_fields = [
                self._create_sub_type(t, f"{self.name}_{i}")
                for i, t in enumerate(get_args(self.type_))
            ]
            self.type_ = origin
            return
        else:
            raise TypeError(f'Fields of type "{origin}" are not supported.')

        # type_ has been refined eg. as the type of a List and sub_fields needs to be populated
        self.sub_fields = [self._create_sub_type(self.type_, "_" + self.name)]

    ModelField._type_analysis = _type_analysis



from __future__ import annotations

from collections.abc import Mapping, Sequence
from functools import cached_property
from typing import Annotated, Any, ClassVar, Literal, Optional, get_args, get_origin

from pydantic import BaseModel, Extra, root_validator, validator
from pydantic.fields import ModelField

from arti.internal.patches import patch_pydantic_ModelField__type_analysis
from arti.internal.type_hints import is_union, lenient_issubclass
from arti.internal.utils import class_name, classproperty, frozendict

patch_pydantic_ModelField__type_analysis()


def _check_types(value: Any, type_: type) -> Any:  # noqa: C901
    mismatch_error = ValueError(f"expected an instance of {type_}, got: {value}")

    if type_ is Any:
        return value
    origin = get_origin(type_)
    if origin is not None:
        args = get_args(type_)
        if origin is Annotated:
            return _check_types(value, args[0])
        if origin is Literal:
            return _check_types(value, type(args[0]))
        # NOTE: Optional[t] -> Union[t, NoneType]
        if is_union(origin):
            for subtype in args:
                try:
                    return _check_types(value, subtype)
                except ValueError:
                    pass
            raise mismatch_error
        if issubclass(origin, (dict, Mapping)):
            value = _check_types(value, origin)
            for k, v in value.items():
                _check_types(k, args[0])
                _check_types(v, args[1])
            return value
        # Variadic tuples will be handled below
        if issubclass(origin, tuple) and ... not in args:
            value = _check_types(value, origin)
            if len(value) != len(args):
                raise mismatch_error
            for i, subtype in enumerate(args):
                _check_types(value[i], subtype)
            return value
        for t in (tuple, list, set, frozenset, Sequence):
            if issubclass(origin, t):
                value = _check_types(value, origin)
                for subvalue in value:
                    _check_types(subvalue, args[0])
                return value
        if set(args) == {Any}:
            return _check_types(value, origin)
        raise NotImplementedError(f"Missing handler for {type_} with {value}!")
    if isinstance(value, Mapping) and not isinstance(value, frozendict):
        value = frozendict(value)
    if not lenient_issubclass(type(value), type_):
        raise mismatch_error
    return value


class Model(BaseModel):
    # A model can be marked _abstract_ to prevent direct instantiation, such as when it is intended
    # as a base class for other models with arbitrary data. As the subclasses of an _abstract_ model
    # have unknown fields (varying per subclass), we don't have targets to mark abstract with
    # abc.ABC nor typing.Protocol. See [1] for more context.
    #
    # 1: https://github.com/replicahq/artigraph/pull/60#discussion_r669089086
    _abstract_: ClassVar[bool] = True
    _class_key_: ClassVar[str] = class_name()

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)  # type: ignore # https://github.com/python/mypy/issues/4660
        # Default _abstract_ to False if not set explicitly on the class. __dict__ is read-only.
        setattr(cls, "_abstract_", cls.__dict__.get("_abstract_", False))

    @root_validator(pre=True)
    @classmethod
    def _block_abstract_instance(cls, values: dict[str, Any]) -> dict[str, Any]:
        if cls._abstract_:
            raise ValueError(f"{cls} cannot be instantiated directly!")
        return values

    @validator("*", pre=True)
    @classmethod
    def _strict_types(cls, value: Any, field: ModelField) -> Any:
        """Check that the value is a stricter instance of the declared type annotation.

        Pydantic will attempt to *parse* values (eg: "5" -> 5), but we'd prefer stricter values for
        clarity and to avoid silent precision loss (eg: 5.3 -> 5).

        NOTE: (at least) one edge case exists with `Union[str, float]` style annotations [1] that
        causes a `5.0` input value to be output as `"5.0"`. This can be worked around by ordering
        the types in the Union from most to least specific (eg: `Union[float, str]`). Alternatively,
        there is an `each_item=True` arg to `@validator` that would let us validate/pick individual
        Union member with `_check_types`, but this mode doesn't let us validate dictionary keys.
        Giving priority to dicts in this case, as I expect they'll be more common.

        1: https://github.com/samuelcolvin/pydantic/issues/1423
        """
        # `field.type_` points to the *inner* type (eg: `int`->`int`; `tuple[int, ...]` -> `int`)
        # while `field.outer_type_` will (mostly) include the full spec and match the `value` we
        # received. The caveat is the `field.outer_type_` will never be wrapped in `Optional`
        # (though nested fields like `tuple[tuple[Optional[int]]]` would). Hence, we pull the
        # `field.outer_type_`, but add back the `Optional` wrapping if necessary.
        type_ = field.outer_type_
        if field.allow_none:
            type_ = Optional[type_]
        return _check_types(value, type_)

    # By default, pydantic just compares models by their dict representation, causing models of
    # different types but same fields (eg: Int8 and Int16) to be equivalent. This can be removed if
    # [1] is merged+released.
    #
    # 1: https://github.com/samuelcolvin/pydantic/pull/3066
    def __eq__(self, other: Any) -> bool:
        return self.__class__ == other.__class__ and tuple(self._iter()) == tuple(other._iter())

    # Omitting unpassed args in repr by default
    def __repr_args__(self) -> Sequence[tuple[Optional[str], Any]]:
        return [
            (k, v)
            for k, v in super().__repr_args__()
            if k in self.__fields_set__
            # NOTE: This check won't be necessary once [1] is released.
            #
            # 1: https://github.com/samuelcolvin/pydantic/pull/2593
            and (True if k is None else self.__fields__[k].field_info.extra.get("repr", True))
        ]

    def __str__(self) -> str:
        return repr(self)

    class Config:
        extra = Extra.forbid
        frozen = True
        keep_untouched = (cached_property, classproperty)
        validate_assignment = True  # Unused with frozen, unless that is overridden in subclass.

    @classmethod
    def _pydantic_type_system_post_field_conversion_hook_(
        cls, type_: arti.types.Type, *, name: str, required: bool
    ) -> arti.types.Type:
        return type_


import arti.types  # noqa: E402 # # pylint: disable=wrong-import-position



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from itertools import chain
from typing import TYPE_CHECKING, Any, ClassVar, Optional

from pydantic import Field, validator
from pydantic.fields import ModelField

from arti.annotations import Annotation
from arti.formats import Format
from arti.internal.models import Model
from arti.internal.utils import classproperty
from arti.partitions import CompositeKeyTypes, PartitionKey
from arti.storage import InputFingerprints, Storage, StoragePartition
from arti.types import Type

if TYPE_CHECKING:
    from arti.producers import ProducerOutput


class BaseArtifact(Model):
    """A BaseArtifact is the most basic data structure describing data in the Artigraph ecosystem.

    A BaseArtifact is comprised of three key elements:
    - type: spec of the data's structure, such as data types, nullable, etc.
    - format: the data's serialized format, such as CSV, Parquet, database native, etc.
    - storage: the data's persistent storage system, such as blob storage, database native, etc.
    """

    # Type *must* be set on the class and be rather static - small additions may be necessary at
    # Graph level (eg: dynamic column additions), but these should be minor. We might allow Struct
    # Types to be "open" (partial type) or "closed".
    #
    # Format and storage *should* be set with defaults on Artifact subclasses to ease most Graph
    # definitions, but will often need to be overridden at the Graph level.
    #
    # In order to override on the instance, avoid ClassVars lest mypy complains when/if we override.
    type: Type
    format: Format
    storage: Storage[Any]

    # Hide in repr to prevent showing the entire upstream graph
    producer_output: Optional[ProducerOutput] = Field(None, repr=False)

    # Class level alias for `type`, which must be set on (non-abstract) subclasses.
    #
    # Pydantic removes class defaults and stashes them in cls.__fields__. To ease access, we
    # automatically populate this from `type` in `__init_subclass__`.
    _type: ClassVar[Type]

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if not cls._abstract_ and cls.__fields__["type"].default is None:
            raise ValueError(f"{cls.__name__} must set `type`")
        cls._type = cls.__fields__["type"].default

    @validator("type", always=True)
    @classmethod
    def _validate_type(cls, type_: Type) -> Type:
        if type_ != cls._type:
            # NOTE: We do a lot of class level validation (particularly in Producer) that relies on
            # the *class* type, such as partition key validation. It's possible we could loosen this
            # a bit by allowing *new* Struct fields, but still requiring an exact match for other
            # Types (including existing Struct fields).
            raise ValueError("overriding `type` is not supported")
        return type_

    @validator("format", always=True)
    @classmethod
    def _validate_format(cls, format: Format, values: dict[str, Any]) -> Format:
        if "type" in values:
            format.supports(type_=values["type"])
        return format

    @validator("storage", always=True)
    @classmethod
    def _validate_storage(cls, storage: Storage[Any], values: dict[str, Any]) -> Storage[Any]:
        # Perhaps I can:
        # - add Artifact.tags (or whatever it is from the Graph)
        # - pass those along to format and storage
        # - change (and rename) .supports to return an updated/copied version w/ new refs or whatever is needed
        #     - eg: so the Storage can save off the PartitionKey.key_types_from(type) and bind the tags to the path(s)
        #
        # The Artifact won't have the tags originally, but on assignment to the Graph, we can copy and
        # update them (probably append/merge w/ existing) and then propagate to the sub-members. I
        # probably don't want to actually store "partition_key_types" on the Storage and maybe not
        # even "tags", so maybe just resolve the tags into the path and validate the left over keys
        # are in the specified partition keys?
        #
        # We'll need to get this validation to run on .copy - or add a .update method[1]. Probably
        # prefer to override the .copy method (and maybe add a _copy that calls the super's .copy in
        # case really needed).
        #
        # 1: https://github.com/samuelcolvin/pydantic/issues/418
        if "type" in values and "format" in values:
            storage.supports(type_=values["type"], format=values["format"])
            storage = storage.resolve_partition_key_spec(cls.partition_key_types).resolve_extension(
                values["format"].extension
            )

        return storage

    @classproperty
    @classmethod
    def partition_key_types(cls) -> CompositeKeyTypes:
        return PartitionKey.types_from(cls._type)

    @classproperty
    @classmethod
    def is_partitioned(cls) -> bool:
        return bool(cls.partition_key_types)

    def discover_storage_partitions(
        self, input_fingerprints: InputFingerprints = InputFingerprints()
    ) -> tuple[StoragePartition, ...]:
        # Hmm, could we lookup the producer and somehow generate the input
        # fingerprints...? Would that be expensive? I guess we kinda want that
        # regardless for any interactive use accessing the graph artifacts. Can we
        # call/run .map and how about resource/backend access?
        return self.storage.discover_partitions(self.partition_key_types, input_fingerprints)


class Statistic(BaseArtifact):
    """A Statistic is a piece of data derived from an Artifact that can be tracked over time."""

    # TODO: Set format/storage to some "system default" that can be used across backends?

    _abstract_ = True


class Artifact(BaseArtifact):
    """An Artifact is the base structure describing an existing or generated dataset.

    An Artifact is comprised of three key elements:
    - `type`: spec of the data's structure, such as data types, nullable, etc.
    - `format`: the data's serialized format, such as CSV, Parquet, database native, etc.
    - `storage`: the data's persistent storage system, such as blob storage, database native, etc.

    In addition to the core elements, an Artifact can be tagged with additional `annotations`
    (to associate it with human knowledge) and `statistics` (to track derived characteristics
    over time).
    """

    _abstract_ = True

    annotations: tuple[Annotation, ...] = ()
    statistics: tuple[Statistic, ...] = ()

    @validator("annotations", "statistics", always=True, pre=True)
    @classmethod
    def _merge_class_defaults(cls, value: tuple[Any, ...], field: ModelField) -> tuple[Any, ...]:
        return tuple(chain(cls.__fields__[field.name].default, value))

    @classmethod
    def cast(cls, value: Any) -> Artifact:
        """Attempt to convert an arbitrary value to an appropriate Artifact instance.

        `Artifact.cast` is used to convert values assigned to an `Artifact.box` (such as
        `Graph.artifacts`) into an Artifact. When called with:
        - an Artifact instance, it is returned
        - a Producer instance with a single output Artifact, the output Artifact is returned
        - a Producer instance with a multiple output Artifacts, an error is raised
        - other types, an error is raised
        """
        from arti.producers import Producer

        # TODO: Leverage a TypeSystem("python") to cast to Artifact classes with "backend native"
        # storage to support builtin assignment and custom type registration.
        if isinstance(value, Artifact):
            return value
        if isinstance(value, Producer):
            output_artifacts = value.out()
            if isinstance(output_artifacts, Artifact):
                return output_artifacts
            n_outputs = len(output_artifacts)
            if n_outputs == 0:  # pragma: no cover
                # TODO: "side effect" Producers: https://github.com/replicahq/artigraph/issues/11
                raise ValueError(f"{type(value).__name__} doesn't produce any Artifacts!")
            assert n_outputs > 1
            raise ValueError(
                f"{type(value).__name__} produces {len(output_artifacts)} Artifacts. Try assigning each to a new name in the Graph!"
            )

        raise NotImplementedError(
            f"Casting python objects ({value}) to Artifacts is not implemented yet!"
        )



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from collections import defaultdict
from collections.abc import Callable, Iterable, Iterator
from inspect import Parameter, Signature, getattr_static
from typing import Annotated, Any, ClassVar, Optional, TypeVar, Union, cast, get_args, get_origin

from toolz import curry

from arti.annotations import Annotation
from arti.artifacts import Artifact, BaseArtifact, Statistic
from arti.fingerprints import Fingerprint
from arti.internal import wrap_exc
from arti.internal.models import Model
from arti.internal.type_hints import NoneType, lenient_issubclass, signature
from arti.internal.utils import classproperty, frozendict, ordinal
from arti.partitions import CompositeKey, CompositeKeyTypes, NotPartitioned
from arti.storage import StoragePartitions
from arti.versions import SemVer, Version
from arti.views import View


def _commas(vals: Iterable[Any]) -> str:
    return ", ".join([str(v) for v in vals])


# TODO: Add @validator over all (build) fields checking for an io.{read,write} handler.


ArtifactViewPair = tuple[type[Artifact], type[View]]
BuildInputViews = frozendict[str, type[View]]
MapInputMetadata = frozendict[str, type[Artifact]]
# mypy doesn't yet support homogeneous tuple aliases: https://github.com/python/mypy/issues/9980
OutputMetadata = tuple[ArtifactViewPair, ...]  # type: ignore

PartitionDependencies = frozendict[CompositeKey, frozendict[str, StoragePartitions]]

BuildSig = Callable[..., Any]
MapSig = Callable[..., PartitionDependencies]

_T = TypeVar("_T")


class Producer(Model):
    """A Producer is a task that builds one or more Artifacts."""

    # User fields/methods

    annotations: tuple[Annotation, ...] = ()
    version: ClassVar[Version] = SemVer(major=0, minor=0, patch=1)

    # build and map must be @classmethods or @staticmethods.
    build: ClassVar[BuildSig]
    map: ClassVar[MapSig]

    # Internal fields/methods

    _abstract_: ClassVar[bool] = True
    # NOTE: The following are set in __init_subclass__
    _input_artifact_types_: ClassVar[frozendict[str, type[Artifact]]]
    _build_sig_: ClassVar[Signature]
    _build_input_views_: ClassVar[BuildInputViews]
    _output_metadata_: ClassVar[OutputMetadata]
    _map_sig_: ClassVar[Signature]
    _map_input_metadata_: ClassVar[MapInputMetadata]

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if not cls._abstract_:
            with wrap_exc(ValueError, prefix=cls.__name__):
                cls._input_artifact_types_ = cls._validate_fields()
                with wrap_exc(ValueError, prefix=".build"):
                    (
                        cls._build_sig_,
                        cls._build_input_views_,
                        cls._output_metadata_,
                    ) = cls._validate_build_sig()
                with wrap_exc(ValueError, prefix=".map"):
                    cls._map_sig_, cls._map_input_metadata_ = cls._validate_map_sig()
                cls._validate_no_unused_fields()

    @classmethod
    def _get_artifact_from_annotation(cls, annotation: Any) -> type[Artifact]:
        hint_example = "eg: `Annotated[pd.DataFrame, MyArtifact]`"
        origin, args = get_origin(annotation), get_args(annotation)
        if origin is not Annotated:
            raise ValueError(f"must be an Annotated hint ({hint_example})")
        # We don't need the underlying type here - `_get_view_from_annotation` requires the entire
        # (possibly Annotated) annotation in order to discover explicitly set Views.
        _, *hints = args
        artifacts = [hint for hint in hints if lenient_issubclass(hint, Artifact)]
        if len(artifacts) == 0:
            raise ValueError(f"Artifact is not set ({hint_example})")
        if len(artifacts) > 1:
            raise ValueError("multiple Artifacts set")
        return cast(type[Artifact], artifacts[0])

    @classmethod
    def _get_view_from_annotation(cls, annotation: Any, artifact: type[Artifact]) -> type[View]:
        view = View.get_class_for(annotation)
        wrap_msg = f"{artifact.__name__}"
        if artifact.is_partitioned:
            wrap_msg = f"partitions of {artifact.__name__}"
        with wrap_exc(ValueError, prefix=f" ({wrap_msg})"):
            view.type_system.check_similarity(arti=artifact._type, python_type=annotation)
        return view

    @classmethod
    def _validate_fields(cls) -> frozendict[str, type[Artifact]]:
        # NOTE: Aside from the base producer fields, all others should be Artifacts.
        #
        # Users can set additional class attributes, but they must be properly hinted as ClassVars.
        # These won't interact with the "framework" and can't be parameters to build/map.
        artifact_fields = {k: v for k, v in cls.__fields__.items() if k not in Producer.__fields__}
        for name, field in artifact_fields.items():
            with wrap_exc(ValueError, prefix=f".{name}"):
                if not (field.default is None and field.default_factory is None and field.required):
                    raise ValueError("field must not have a default nor be Optional.")
                if not lenient_issubclass(field.outer_type_, Artifact):
                    raise ValueError(
                        f"type hint must be an Artifact subclass, got: {field.outer_type_}"
                    )
        return frozendict({name: field.outer_type_ for name, field in artifact_fields.items()})

    @classmethod
    def _validate_parameters(
        cls, sig: Signature, *, validator: Callable[[str, Parameter, type[Artifact]], _T]
    ) -> Iterator[_T]:
        if undefined_params := set(sig.parameters) - set(cls._input_artifact_types_):
            raise ValueError(
                f"the following parameter(s) must be defined as a field: {undefined_params}"
            )
        for name, param in sig.parameters.items():
            with wrap_exc(ValueError, prefix=f" {name} param"):
                if param.annotation is param.empty:
                    raise ValueError("must have a type hint.")
                if param.default is not param.empty:
                    raise ValueError("must not have a default.")
                if param.kind not in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY):
                    raise ValueError("must be usable as a keyword argument.")
                artifact = cls.__fields__[param.name].outer_type_
                yield validator(name, param, artifact)

    @classmethod
    def _validate_build_sig_return(cls, annotation: Any, *, i: int) -> ArtifactViewPair:
        with wrap_exc(ValueError, prefix=f" {ordinal(i)} return"):
            artifact = cls._get_artifact_from_annotation(annotation)
            return artifact, cls._get_view_from_annotation(annotation, artifact)

    @classmethod
    def _validate_build_sig(cls) -> tuple[Signature, BuildInputViews, OutputMetadata]:
        """Validate the .build method"""
        if not hasattr(cls, "build"):
            raise ValueError("must be implemented")
        if not isinstance(getattr_static(cls, "build"), (classmethod, staticmethod)):
            raise ValueError("must be a @classmethod or @staticmethod")
        build_sig = signature(cls.build, force_tuple_return=True, remove_owner=True)
        # Validate the parameters
        build_input_metadata = BuildInputViews(
            cls._validate_parameters(
                build_sig,
                validator=(
                    lambda name, param, artifact: (
                        name,
                        cls._get_view_from_annotation(param.annotation, artifact),
                    )
                ),
            )
        )
        # Validate the return definition
        return_annotation = build_sig.return_annotation
        if return_annotation is build_sig.empty:
            # TODO: "side effect" Producers: https://github.com/replicahq/artigraph/issues/11
            raise ValueError("a return value must be set with the output Artifact(s).")
        if return_annotation == (NoneType,):
            raise ValueError("missing return signature")
        output_metadata = OutputMetadata(
            cls._validate_build_sig_return(annotation, i=i)
            for i, annotation in enumerate(return_annotation, 1)
        )
        # Validate all output Artifacts have equivalent partitioning schemes.
        #
        # We currently require the partition key type *and* name to match, but in the
        # future we might be able to extend the dependency metadata to support
        # heterogeneous names if necessary.
        artifacts_by_composite_key = defaultdict[CompositeKeyTypes, list[type[Artifact]]](list)
        for (artifact, _) in output_metadata:
            artifacts_by_composite_key[artifact.partition_key_types].append(artifact)
        if len(artifacts_by_composite_key) != 1:
            raise ValueError("all output Artifacts must have the same partitioning scheme")
        # TODO: Save off output composite_key_types

        return build_sig, build_input_metadata, output_metadata

    @classmethod
    def _validate_map_sig(cls) -> tuple[Signature, MapInputMetadata]:
        """Validate partitioned Artifacts and the .map method"""
        if not hasattr(cls, "map"):
            partitioned_outputs = [
                artifact for (artifact, view) in cls._output_metadata_ if artifact.is_partitioned
            ]
            # TODO: Add runtime checking of `map` output (ie: output aligns w/ output
            # artifacts and such).
            if partitioned_outputs:
                raise ValueError("must be implemented when the `build` outputs are partitioned")
            else:

                def map(**kwargs: StoragePartitions) -> PartitionDependencies:
                    return PartitionDependencies(
                        {NotPartitioned: {name: partitions for name, partitions in kwargs.items()}}
                    )

            # Narrow the map signature, which is validated below and used at graph build
            # time (via cls._map_input_metadata_) to determine what arguments to pass to
            # map.
            map.__signature__ = Signature(  # type: ignore
                [
                    Parameter(name=name, annotation=StoragePartitions, kind=Parameter.KEYWORD_ONLY)
                    for name, artifact in cls._input_artifact_types_.items()
                    if name in cls._build_input_views_
                ],
                return_annotation=PartitionDependencies,
            )
            cls.map = staticmethod(map)  # type: ignore
        if not isinstance(getattr_static(cls, "map"), (classmethod, staticmethod)):
            raise ValueError("must be a @classmethod or @staticmethod")
        map_sig = signature(cls.map)

        def validate_map_param(
            name: str, param: Parameter, artifact: type[Artifact]
        ) -> tuple[str, type[Artifact]]:
            # TODO: Should we add some ArtifactPartition[MyArtifact] type?
            if param.annotation != StoragePartitions:
                raise ValueError("type hint must be `StoragePartitions`")
            return name, artifact

        map_input_metadata = MapInputMetadata(
            cls._validate_parameters(map_sig, validator=validate_map_param)
        )
        return map_sig, map_input_metadata  # TODO: Verify map output hint matches TBD spec

    @classmethod
    def _validate_no_unused_fields(cls) -> None:
        if unused_fields := set(cls._input_artifact_types_) - (
            set(cls._build_sig_.parameters) | set(cls._map_sig_.parameters)
        ):
            raise ValueError(
                f"the following fields aren't used in `.build` or `.map`: {unused_fields}"
            )

    # NOTE: pydantic defines .__iter__ to return `self.__dict__.items()` to support `dict(model)`,
    # but we want to override to support easy expansion/assignment to a Graph  without `.out()` (eg:
    # `g.artifacts.a, g.artifacts.b = MyProducer(...)`).
    def __iter__(self) -> Iterator[Artifact]:  # type: ignore
        ret = self.out()
        if not isinstance(ret, tuple):
            ret = (ret,)
        return iter(ret)

    @classproperty
    @classmethod
    def fingerprint(cls) -> Fingerprint:
        """Return a Fingerprint of the Producer key + version.

        The input and output Artifacts are ignored as the Producer instance may be used multiple
        times to produce different output partitions. The "entropy mixing" will be performed for
        *each* output with the static Producer.fingerprint + that output's specific input
        partition dependencies.
        """
        return Fingerprint.from_string(cls._class_key_).combine(cls.version.fingerprint)

    def compute_input_fingerprint(
        self, dependency_partitions: frozendict[str, StoragePartitions]
    ) -> Fingerprint:
        return self.fingerprint.combine(
            *(
                partition.fingerprint
                for name, partitions in dependency_partitions.items()
                for partition in partitions
            )
        )
        # partition_input_fingerprints = {
        #     composite_key:
        #     for composite_key, dependency_partitions in partition_dependencies.items()
        # }

    @property
    def inputs(self) -> dict[str, Artifact]:
        return {k: getattr(self, k) for k in self._input_artifact_types_}

    def out(self, *outputs: Artifact) -> Union[Artifact, tuple[Artifact, ...]]:
        """Configure the output Artifacts this Producer will build.

        The arguments are matched to the `Producer.build` return signature in order.
        """
        if not outputs:
            # TODO: Raise a better error if the Artifacts don't have defaults set for
            # type/format/storage.
            outputs = tuple(artifact() for (artifact, _) in self._output_metadata_)
        passed_n, expected_n = len(outputs), len(self._build_sig_.return_annotation)
        if passed_n != expected_n:
            ret_str = _commas(self._build_sig_.return_annotation)
            raise ValueError(
                f"{self._class_key_}.out() - expected {expected_n} arguments of ({ret_str}), but got: {outputs}"
            )

        def validate(artifact: Artifact, *, ord: int) -> Artifact:
            (expected_type, _) = self._output_metadata_[ord]
            with wrap_exc(ValueError, prefix=f"{self._class_key_}.out() {ordinal(ord+1)} argument"):
                if not isinstance(artifact, expected_type):
                    raise ValueError(f"expected instance of {expected_type}, got {type(artifact)}")
                # TODO: Validate type/format/storage/view compatibility?
                if artifact.producer_output is not None:
                    raise ValueError(
                        f"{artifact} is produced by {artifact.producer_output.producer}!"
                    )
            return artifact.copy(
                update={"producer_output": ProducerOutput(producer=self, position=ord)}
            )

        outputs = tuple(validate(artifact, ord=i) for i, artifact in enumerate(outputs))
        if len(outputs) == 1:
            return outputs[0]
        return outputs


@curry
def producer(
    build: BuildSig,
    *,
    annotations: Optional[tuple[Annotation, ...]] = None,
    map: Optional[MapSig] = None,
    name: Optional[str] = None,
    version: Optional[Version] = None,
) -> type[Producer]:
    name = build.__name__ if name is None else name
    input_artifacts = {}
    for param in signature(build).parameters.values():
        with wrap_exc(ValueError, prefix=f"{name} {param.name} param"):
            input_artifacts[param.name] = Producer._get_artifact_from_annotation(param.annotation)
    return type(
        build.__name__,
        (Producer,),
        {
            k: v
            for k, v in {
                "__annotations__": input_artifacts,
                "annotations": annotations,
                "build": staticmethod(build),
                "map": None if map is None else staticmethod(map),
                "version": version,
            }.items()
            if v is not None
        },
    )


class ProducerOutput(Model):
    producer: Producer
    position: int  # TODO: Support named output (defaulting to artifact classname)


BaseArtifact.update_forward_refs(ProducerOutput=ProducerOutput)
Statistic.update_forward_refs(ProducerOutput=ProducerOutput)
Artifact.update_forward_refs(ProducerOutput=ProducerOutput)



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from typing import Any, ClassVar

from arti.types import Type


class Threshold:
    type: ClassVar[type[Type]]

    def check(self, value: Any) -> bool:
        raise NotImplementedError()



from __future__ import annotations

import string
from collections import defaultdict
from collections.abc import Callable, Mapping
from functools import partial
from typing import Any, Iterable, Literal, Optional, overload

import parse

from arti.fingerprints import Fingerprint
from arti.internal.utils import frozendict
from arti.partitions import CompositeKey, CompositeKeyTypes, PartitionKey

InputFingerprints = frozendict[CompositeKey, Fingerprint]


class Placeholder:
    def __init__(self, key: str) -> None:
        self._key = key


class FormatPlaceholder(Placeholder):
    def __format__(self, spec: str) -> str:
        result = self._key
        if spec:
            result += ":" + spec
        return "{" + result + "}"

    def __getitem__(self, key: Any) -> FormatPlaceholder:
        self._key = f"{self._key}[{key}]"
        return self

    def __getattr__(self, attr: str) -> FormatPlaceholder:
        self._key = f"{self._key}.{attr}"
        return self


class WildcardPlaceholder(Placeholder):
    def __init__(self, key: str, key_types: CompositeKeyTypes):
        super().__init__(key)
        if key not in key_types:
            raise ValueError(f"No '{key}' partition key found, expected one of {tuple(key_types)}")
        self._key_type = key_types[key]
        self._attribute: Optional[str] = None  # What field/key_component is being accessed

    @classmethod
    def with_key_types(
        cls, key_types: Mapping[str, type[PartitionKey]]
    ) -> Callable[[str], WildcardPlaceholder]:
        return partial(cls, key_types=key_types)

    def _err_if_no_attribute(self) -> None:
        if self._attribute is None:
            example = sorted(self._key_type.key_components)[0]
            raise ValueError(
                f"'{self._key}' cannot be used directly in a partition path; access one of the key components (eg: '{self._key}.{example}')."
            )

    def __getattr__(self, name: str) -> WildcardPlaceholder:
        if self._attribute is not None:
            raise ValueError(
                f"'{self._key}.{self._attribute}.{name}' cannot be used in a partition path; only immediate '{self._key}' attributes (such as '{self._attribute}') can be used."
            )
        if name not in self._key_type.key_components:
            raise AttributeError(
                f"'{self._key_type.__name__}' has no field or key component '{name}'"
            )
        self._attribute = name
        return self

    def __getitem__(self, key: Any) -> Any:
        self._err_if_no_attribute()
        # Pass through "hard coded" partition parts, eg: "{date.Y:2021}" -> format_spec="2021".
        #
        # TODO: Should we validate the key against the key_type/attribute type?
        return key

    def __str__(self) -> str:
        self._err_if_no_attribute()
        return "*"


# NOTE: We might pre-populate this with the Graph tags so the hard coded templates are already available (rather than
# being filled in with a WildcardPlaceholder).
class FormatDict(dict[str, Any]):
    def __init__(
        self,
        placeholder_type: Callable[[str], Placeholder],
        /,
        *args: Iterable[tuple[str, Any]],
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.placeholder_type = placeholder_type

    def __missing__(self, key: str) -> Placeholder:
        return self.placeholder_type(key)


def partial_format(spec: str, **kwargs: Any) -> str:
    return string.Formatter().vformat(spec, (), FormatDict(FormatPlaceholder, **kwargs))


def spec_to_wildcard(spec: str, key_types: Mapping[str, type[PartitionKey]]) -> str:
    return string.Formatter().vformat(
        spec.replace("{input_fingerprint}", "*"),
        (),
        FormatDict(WildcardPlaceholder.with_key_types(key_types)),
    )


@overload
def extract_placeholders(
    *,
    error_on_no_match: Literal[False],
    key_types: Mapping[str, type[PartitionKey]],
    parser: parse.Parser,
    path: str,
    spec: str,
) -> tuple[Optional[Fingerprint], Optional[CompositeKey]]:
    ...


@overload
def extract_placeholders(
    *,
    error_on_no_match: Literal[True] = True,
    key_types: Mapping[str, type[PartitionKey]],
    parser: parse.Parser,
    path: str,
    spec: str,
) -> tuple[Optional[Fingerprint], CompositeKey]:
    ...


def extract_placeholders(
    *,
    error_on_no_match: bool = True,
    key_types: Mapping[str, type[PartitionKey]],
    parser: parse.Parser,
    path: str,
    spec: str,
) -> tuple[Optional[Fingerprint], Optional[CompositeKey]]:
    parsed_value = parser.parse(path)
    if parsed_value is None:
        if error_on_no_match:
            raise ValueError(f"Unable to parse '{path}' with '{spec}'.")
        return None, None
    input_fingerprint = None
    key_components = defaultdict[str, dict[str, str]](dict)
    for k, v in parsed_value.named.items():
        if k == "input_fingerprint":
            assert isinstance(v, str)
            input_fingerprint = Fingerprint.from_int(int(v))
            continue
        key, component = k.split(".")
        # parsing a string like "{date.Y[1970]}" will return a dict like {'1970': '1970'}.
        if isinstance(v, dict):
            v = tuple(v)[0]
        key_components[key][component] = v

    keys = {
        key: key_types[key].from_key_components(**components)
        for key, components in key_components.items()
    }
    if set(keys) != set(key_types):
        raise ValueError(
            f"Expected to find partition keys for {sorted(key_types)}, only found {sorted(keys)}. Is the partitioning spec ('{spec}') complete?"
        )
    return input_fingerprint, frozendict(keys)


def parse_spec(
    paths: set[str],
    *,
    spec: str,
    key_types: Mapping[str, type[PartitionKey]],
    input_fingerprints: InputFingerprints,
) -> Mapping[str, tuple[Optional[Fingerprint], CompositeKey]]:
    parser = parse.compile(spec, case_sensitive=True)
    return {
        path: (input_fingerprint, keys)
        for (path, (input_fingerprint, keys)) in (
            (path, extract_placeholders(parser=parser, path=path, spec=spec, key_types=key_types))
            for path in paths
        )
        if input_fingerprints.get(keys) == input_fingerprint
    }



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

import inspect
import subprocess
from abc import abstractmethod
from collections.abc import Callable
from datetime import datetime, timezone
from typing import Any, cast

from pydantic import Field, validator

from arti.fingerprints import Fingerprint
from arti.internal.models import Model
from arti.internal.utils import qname


class Version(Model):
    _abstract_ = True

    @property
    @abstractmethod
    def fingerprint(self) -> Fingerprint:
        raise NotImplementedError(f"{qname(self)}.fingerprint is not implemented!")


class GitCommit(Version):
    sha: str = Field(
        default_factory=(
            lambda: subprocess.check_output(["git", "rev-parse", "HEAD"]).decode().strip()
        )
    )

    @property
    def fingerprint(self) -> Fingerprint:
        return Fingerprint.from_string(self.sha)


class SemVer(Version):
    """SemVer fingerprinting only considers the major component, unless it is less than 0.

    By only considering the major version, we can add incremental bumps to a Producer without triggering
    historical backfills. The major version MUST be incremented on schema or methodological changes.
    """

    major: int
    minor: int
    patch: int

    @property
    def fingerprint(self) -> Fingerprint:
        s = str(self.major)
        if self.major == 0:
            s = f"{self.major}.{self.minor}.{self.patch}"
        return Fingerprint.from_string(s)


class String(Version):
    value: str

    @property
    def fingerprint(self) -> Fingerprint:
        return Fingerprint.from_string(self.value)


class _SourceDescriptor:  # Experimental :)
    # Using AST rather than literal source will likely be less "noisy":
    #     https://github.com/replicahq/artigraph/pull/36#issuecomment-824131156
    def __get__(self, obj: Any, type_: type) -> String:
        return String(value=inspect.getsource(type_))


_Source = cast(Callable[[], String], _SourceDescriptor)


class Timestamp(Version):
    dt: datetime = Field(default_factory=lambda: datetime.now(tz=timezone.utc))

    @validator("dt", always=True)
    @classmethod
    def _requires_timezone(cls, dt: datetime) -> datetime:
        if dt is not None and dt.tzinfo is None:
            raise ValueError("Timestamp requires a timezone-aware datetime!")
        return dt

    @property
    def fingerprint(self) -> Fingerprint:
        return Fingerprint.from_int(round((self.dt or datetime.utcnow()).timestamp()))


# TODO: Consider a Timestamp like version with a "frequency" arg (day, hour, etc) that we floor/ceil
# to trigger "scheduled" invalidation. This doesn't solve imperative work like "process X every hour
# on the hour and play catch up for missed runs" (rather, we should aim to represent that with
# upstream Artifact partitions), but rather solves work like "ingest all data at most this
# frequently" (eg: full export from an API - no point in a backfill).



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from collections.abc import Iterator
from contextlib import contextmanager
from typing import Protocol

from arti.artifacts import Artifact
from arti.internal.models import Model
from arti.storage import StoragePartitions


class BackendProtocol(Protocol):
    def read_artifact_partitions(self, artifact: Artifact) -> StoragePartitions:
        raise NotImplementedError()

    def write_artifact_partitions(self, artifact: Artifact, partitions: StoragePartitions) -> None:
        raise NotImplementedError()


class Backend(Model):
    """Backend represents a storage for internal Artigraph metadata.

    Backend storage is an addressable location (local path, database connection, etc) that
    tracks metadata for a collection of Graphs over time, including:
    - the Artifact(s)->Producer->Artifact(s) dependency graph
    - Artifact Annotations, Statistics, Partitions, and other metadata
    - Artifact and Producer Fingerprints
    - etc
    """

    @contextmanager
    def connect(self) -> Iterator[BackendProtocol]:
        raise NotImplementedError()



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from functools import reduce
from operator import xor
from typing import Optional

import farmhash

from arti.internal.models import Model
from arti.internal.utils import int64, uint64


class Fingerprint(Model):
    """Fingerprint represents a unique identity as an int64 value.

    Using an int(64) has a number of convenient properties:
    - can be combined independent of order with XOR
    - can be stored relatively cheaply
    - empty 0 values drop out when combined (5 ^ 0 = 5)
    - is relatively cross-platform (across databases, languages, etc)

    There are two "special" Fingerprints w/ factory functions that, when combined with other
    Fingerprints:
    - `empty()`: returns `empty()`
    - `identity()`: return the other Fingerprint
    """

    key: Optional[int64]

    def combine(self, *others: Fingerprint) -> Fingerprint:
        return reduce(xor, others, self)

    @classmethod
    def empty(cls) -> Fingerprint:
        """Return a Fingerprint that, when combined, will return Fingerprint.empty()"""
        return cls(key=None)

    @classmethod
    def from_int(cls, x: int, /) -> Fingerprint:
        return cls.from_int64(int64(x))

    @classmethod
    def from_int64(cls, x: int64, /) -> Fingerprint:
        return cls(key=x)

    @classmethod
    def from_string(cls, x: str, /) -> Fingerprint:
        """Fingerprint an arbitrary string.

        Fingerprints using Farmhash Fingerprint64, converted to int64 via two's complement.
        """
        return cls.from_uint64(uint64(farmhash.fingerprint64(x)))

    @classmethod
    def from_uint64(cls, x: uint64, /) -> Fingerprint:
        return cls.from_int64(int64(x))

    @classmethod
    def identity(cls) -> Fingerprint:
        """Return a Fingerprint that, when combined, will return the other Fingerprint."""
        return cls(key=int64(0))

    @property
    def is_empty(self) -> bool:
        return self.key is None

    @property
    def is_identity(self) -> bool:
        return self.key == 0

    def __xor__(self, other: Fingerprint) -> Fingerprint:
        if self.key is None:
            return Fingerprint.empty()
        if isinstance(other, Fingerprint):
            if other.key is None:
                return Fingerprint.empty()
            return Fingerprint(key=self.key ^ other.key)
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, int):
            return self.key == other
        if isinstance(other, Fingerprint):
            return self.key == other.key
        return NotImplemented



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager

from pydantic import Field

from arti.artifacts import Artifact
from arti.backends import Backend
from arti.storage import StoragePartitions


class MemoryBackend(Backend):
    artifact_partitions: dict[Artifact, StoragePartitions] = Field(default_factory=dict, repr=False)

    @contextmanager
    def connect(self) -> Iterator[MemoryBackend]:
        yield self

    def read_artifact_partitions(self, artifact: Artifact) -> StoragePartitions:
        return self.artifact_partitions[artifact]

    def write_artifact_partitions(self, artifact: Artifact, partitions: StoragePartitions) -> None:
        self.artifact_partitions[artifact] = tuple(
            set(self.artifact_partitions.get(artifact, ()) + partitions)
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['arti',
 'arti.annotations',
 'arti.artifacts',
 'arti.backends',
 'arti.executors',
 'arti.fingerprints',
 'arti.formats',
 'arti.graphs',
 'arti.internal',
 'arti.io',
 'arti.partitions',
 'arti.producers',
 'arti.statistics',
 'arti.storage',
 'arti.thresholds',
 'arti.types',
 'arti.versions',
 'arti.views']

package_data = \
{'': ['*']}

install_requires = \
['deprecation>=2.1.0,<3.0.0',
 'frozendict>=2.0.6,<3.0.0',
 'multimethod>=1.6,<2.0',
 'parse>=1.19.0,<2.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'pyfarmhash>=0.2.2,<0.3.0',
 'python-box>=5.4.1,<6.0.0',
 'setuptools>=58.2.0,<59.0.0',
 'sgqlc>=14.1,<15.0',
 'toolz>=0.11.1,<0.12.0']

setup_kwargs = {
    'name': 'arti',
    'version': '0.0.1a0',
    'description': '',
    'long_description': None,
    'author': 'Replica',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

from typing import Annotated, Any, ClassVar, get_args, get_origin

from arti.internal.models import Model
from arti.internal.type_hints import lenient_issubclass
from arti.internal.utils import import_submodules, register
from arti.types import TypeSystem


class View(Model):
    """View represents the in-memory representation of the artifact.

    Examples include pandas.DataFrame, dask.DataFrame, a BigQuery table.
    """

    _abstract_ = True
    _by_python_type_: ClassVar[dict[type, type[View]]] = {}

    priority: ClassVar[int] = 0  # Set priority of this view for its python_type. Higher is better.
    python_type: ClassVar[type]
    type_system: ClassVar[TypeSystem]

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if not cls._abstract_:
            register(cls._by_python_type_, cls.python_type, cls, lambda x: x.priority)

    @classmethod
    def get_class_for(cls, annotation: Any) -> type[View]:
        import_submodules(__path__, __name__)

        origin, args = get_origin(annotation), get_args(annotation)
        if origin is Annotated:
            annotation, *hints = args
            views: list[type[View]] = [hint for hint in hints if lenient_issubclass(hint, View)]
            if len(views) == 0:
                return cls.get_class_for(annotation)
            if len(views) == 1:
                return views[0]
            raise ValueError("multiple Views set")
        if origin is None:
            origin = annotation
        if origin not in cls._by_python_type_:
            raise ValueError(
                f"{annotation} cannot be matched to a View, try setting one explicitly (eg: `Annotated[pd.DataFrame, MyArtifact, PandasDataFrame]`)"
            )
        return cls._by_python_type_[origin]



import json
from itertools import chain
from pathlib import Path
from typing import Any

from arti.formats.json import JSON
from arti.io import read, write
from arti.storage.local import LocalFilePartition
from arti.types import List, Type
from arti.views.python import PythonBuiltin

# TODO: Do I need to inject the partition keys into the returned data? Likely useful...
# Maybe a View option?


def _read_json_file(path: str) -> Any:
    with open(path, "rb") as file:
        return json.load(file)


@read.register
def _read_json_localfile_python(
    type: Type,
    format: JSON,
    storage_partitions: tuple[LocalFilePartition, ...],
    view: PythonBuiltin,
) -> Any:
    if isinstance(type, List) and type.is_partitioned:
        return list(
            chain.from_iterable(
                _read_json_file(storage_partition.path) for storage_partition in storage_partitions
            )
        )
    else:
        assert len(storage_partitions) == 1
        return _read_json_file(storage_partitions[0].path)


@write.register
def _write_json_localfile_python(
    data: Any, type: Type, format: JSON, storage_partition: LocalFilePartition, view: PythonBuiltin
) -> None:
    path = Path(storage_partition.path)
    path.parent.mkdir(exist_ok=True, parents=True)
    with path.open("w") as file:
        json.dump(data, file)



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

import abc
from datetime import date
from inspect import getattr_static
from typing import Any, ClassVar

from arti.internal.models import Model
from arti.internal.utils import classproperty, frozendict, register
from arti.types import Date, Int8, Int16, Int32, Int64, List, Null, Type


class key_component(property):
    pass


class PartitionKey(Model):
    _abstract_ = True
    _by_type_: ClassVar[dict[type[Type], type[PartitionKey]]] = {}

    default_key_components: ClassVar[tuple[str, ...]]
    matching_type: ClassVar[type[Type]]

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if cls._abstract_:
            return
        for attr in ("default_key_components", "matching_type"):
            if not hasattr(cls, attr):
                raise TypeError(f"{cls.__name__} must set `{attr}`")
        if unknown := (set(cls.default_key_components) - cls.key_components):
            raise TypeError(
                f"Unknown key_components in {cls.__name__}.default_key_components: {unknown}"
            )
        register(cls._by_type_, cls.matching_type, cls)

    @classproperty
    @classmethod
    def key_components(cls) -> frozenset[str]:
        return frozenset(cls.__fields__) | frozenset(
            name for name in dir(cls) if isinstance(getattr_static(cls, name), key_component)
        )

    @classmethod
    @abc.abstractmethod
    def from_key_components(cls, **key_components: str) -> PartitionKey:
        raise NotImplementedError(f"Unable to parse '{cls.__name__}' from: {key_components}")

    @classmethod
    def get_class_for(cls, type_: Type) -> type[PartitionKey]:
        return cls._by_type_[type(type_)]

    @classmethod
    def types_from(cls, type_: Type) -> CompositeKeyTypes:
        if not isinstance(type_, List):
            return frozendict()
        return frozendict(
            {name: cls.get_class_for(field) for name, field in type_.partition_fields.items()}
        )


# CompositeKey is the set of named PartitionKeys that uniquely identify a single partition.
CompositeKey = frozendict[str, PartitionKey]
CompositeKeyTypes = frozendict[str, type[PartitionKey]]

NotPartitioned = CompositeKey()


class DateKey(PartitionKey):
    default_key_components = ("Y", "m", "d")
    matching_type = Date

    key: date

    @key_component
    def Y(self) -> int:
        return self.key.year

    @key_component
    def m(self) -> int:
        return self.key.month

    @key_component
    def d(self) -> int:
        return self.key.day

    @key_component
    def iso(self) -> str:
        return self.key.isoformat()

    @classmethod
    def from_key_components(cls, **key_components: str) -> PartitionKey:
        names = set(key_components)
        if names == {"key"}:
            return cls(key=date.fromisoformat(key_components["key"]))
        if names == {"iso"}:
            return cls(key=date.fromisoformat(key_components["iso"]))
        if names == {"Y", "m", "d"}:
            return cls(key=date(*[int(key_components[k]) for k in ("Y", "m", "d")]))
        return super().from_key_components(**key_components)


class _IntKey(PartitionKey):
    _abstract_ = True
    default_key_components = ("key",)

    key: int

    @key_component
    def hex(self) -> str:
        return hex(self.key)

    @classmethod
    def from_key_components(cls, **key_components: str) -> PartitionKey:
        names = set(key_components)
        if names == {"key"}:
            return cls(key=int(key_components["key"]))
        if names == {"hex"}:
            return cls(key=int(key_components["hex"], base=16))
        return super().from_key_components(**key_components)


class Int8Key(_IntKey):
    matching_type = Int8


class Int16Key(_IntKey):
    matching_type = Int16


class Int32Key(_IntKey):
    matching_type = Int32


class Int64Key(_IntKey):
    matching_type = Int64


class NullKey(PartitionKey):
    default_key_components = ("key",)
    matching_type = Null

    key: None = None

    @classmethod
    def from_key_components(cls, **key_components: str) -> PartitionKey:
        if set(key_components) == {"key"}:
            if key_components["key"] != "None":
                raise ValueError(f"'{cls.__name__}' can only be used with 'None'!")
            return cls()
        return super().from_key_components(**key_components)



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore

import threading
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from arti.graphs import Graph


class _Context(threading.local):
    def __init__(self) -> None:
        super().__init__()
        self.graph: Optional[Graph] = None


context = _Context()



import pickle
from pathlib import Path
from typing import Any

from arti.formats.pickle import Pickle
from arti.io import read, write
from arti.storage.local import LocalFilePartition
from arti.types import Type
from arti.views.python import PythonBuiltin


def _read_pickle_file(path: str) -> Any:
    with open(path, "rb") as file:
        return pickle.load(file)


# TODO: Need to handle partitioned lists. Should I separate read and read_partitions?


@read.register
def _read_pickle_localfile_python(
    type: Type,
    format: Pickle,
    storage_partitions: tuple[LocalFilePartition, ...],
    view: PythonBuiltin,
) -> Any:
    return [_read_pickle_file(storage_partition.path) for storage_partition in storage_partitions]


@write.register
def _write_pickle_localfile_python(
    data: Any,
    type: Type,
    format: Pickle,
    storage_partition: LocalFilePartition,
    view: PythonBuiltin,
) -> None:
    path = Path(storage_partition.path)
    path.parent.mkdir(exist_ok=True, parents=True)
    with path.open("wb") as file:
        pickle.dump(data, file)
